# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from composer.callbacks import (
    EarlyStopper,
    Generate,
    LoadCheckpoint,
    LRMonitor,
    MemoryMonitor,
    MemorySnapshot,
    NaNMonitor,
    OOMObserver,
    OptimizerMonitor,
    RuntimeEstimator,
    SpeedMonitor,
    SystemMetricsMonitor,
)

from llmfoundry.callbacks.async_eval_callback import AsyncEval
from llmfoundry.callbacks.curriculum_learning_callback import CurriculumLearning
from llmfoundry.callbacks.dataset_swap_callback import DatasetSwap
from llmfoundry.callbacks.env_logging_callback import EnvironmentLoggingCallback
from llmfoundry.callbacks.eval_gauntlet_callback import EvalGauntlet
from llmfoundry.callbacks.eval_output_logging_callback import EvalOutputLogging
from llmfoundry.callbacks.fdiff_callback import FDiffMetrics
from llmfoundry.callbacks.hf_checkpointer import HuggingFaceCheckpointer
from llmfoundry.callbacks.kill_loss_spike_callback import KillLossSpike
from llmfoundry.callbacks.log_mbmoe_tok_per_expert_callback import (
    MegaBlocksMoE_TokPerExpert,
)
from llmfoundry.callbacks.loss_perp_v_len_callback import \
    LossPerpVsContextLengthLogger
from llmfoundry.callbacks.monolithic_ckpt_callback import (
    MonolithicCheckpointSaver,
)
from llmfoundry.callbacks.resumption_callbacks import (
    GlobalLRScaling,
    LayerFreezing,
)
from llmfoundry.callbacks.run_timeout_callback import RunTimeoutCallback
from llmfoundry.callbacks.scheduled_gc_callback import ScheduledGarbageCollector
from llmfoundry.registry import callbacks, callbacks_with_config

callbacks.register('system_metrics_monitor', func=SystemMetricsMonitor)
callbacks.register('lr_monitor', func=LRMonitor)
callbacks.register('memory_monitor', func=MemoryMonitor)
callbacks.register('memory_snapshot', func=MemorySnapshot)
callbacks.register('speed_monitor', func=SpeedMonitor)
callbacks.register('runtime_estimator', func=RuntimeEstimator)
callbacks.register('optimizer_monitor', func=OptimizerMonitor)
callbacks.register('generate_callback', func=Generate)
callbacks.register('early_stopper', func=EarlyStopper)
callbacks.register('fdiff_metrics', func=FDiffMetrics)
callbacks.register('hf_checkpointer', func=HuggingFaceCheckpointer)
callbacks.register('global_lr_scaling', func=GlobalLRScaling)
callbacks.register('layer_freezing', func=LayerFreezing)
callbacks.register('mono_checkpoint_saver', func=MonolithicCheckpointSaver)
callbacks.register('scheduled_gc', func=ScheduledGarbageCollector)
callbacks.register('oom_observer', func=OOMObserver)
callbacks.register('eval_output_logging', func=EvalOutputLogging)
callbacks.register('mbmoe_tok_per_expert', func=MegaBlocksMoE_TokPerExpert)
callbacks.register('run_timeout', func=RunTimeoutCallback)
callbacks.register('loss_perp_v_len', func=LossPerpVsContextLengthLogger)
callbacks.register('env_logging', func=EnvironmentLoggingCallback)
callbacks.register('nan_monitor', func=NaNMonitor)
callbacks.register('kill_loss_spike', func=KillLossSpike)
callbacks.register('load_checkpoint', func=LoadCheckpoint)

callbacks_with_config.register('async_eval', func=AsyncEval)
callbacks_with_config.register('curriculum_learning', func=CurriculumLearning)
callbacks_with_config.register('dataset_swap', func=DatasetSwap)

__all__ = [
    'FDiffMetrics',
    'MonolithicCheckpointSaver',
    'GlobalLRScaling',
    'LayerFreezing',
    'ScheduledGarbageCollector',
    'EvalGauntlet',
    'HuggingFaceCheckpointer',
    'MegaBlocksMoE_TokPerExpert',
    'AsyncEval',
    'CurriculumLearning',
    'LossPerpVsContextLengthLogger',
    'KillLossSpike',
]



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any, Callable, Iterable, Mapping, Union

import torch
import transformers
from composer.core.data_spec import DataSpec
from composer.core.types import Batch
from torch.utils.data import DataLoader as TorchDataloader
from transformers import PreTrainedTokenizerBase

from llmfoundry.data.finetuning.collator import Seq2SeqFinetuningCollator
from llmfoundry.data.finetuning.dataloader import build_collate_fn
from llmfoundry.data.packing import BinPackCollator
from llmfoundry.data.text_data import ConcatenatedSequenceCollatorWrapper
from llmfoundry.utils.consts import CROSS_ENTROPY_IGNORE_INDEX

log = logging.getLogger(__name__)


class LossGeneratingTokensCollatorWrapper:
    """Collator wrapper to add loss generating token counts to batch."""

    def __init__(
        self,
        base_collator: Callable,
        token_counting_func: Callable[[Batch], Union[int, dict[str, int]]],
    ):
        self.base_collator = base_collator
        self.token_counting_func = token_counting_func

        self._token_count_batch_keys = [
            'input_ids',
            'attention_mask',
            'labels',
            'decoder_attention_mask',
        ]

    def __call__(self, examples: list[Any]) -> dict[str, torch.Tensor]:
        batch = self.base_collator(examples)

        # Add token counts to batch as a list, one for each row, so that microbatch splitting works
        output = {
            'total_tokens': [],
            'loss_generating_tokens': [],
        }
        num_rows = batch['input_ids'].shape[0]
        for row in range(num_rows):
            row_batch = {}
            for key in self._token_count_batch_keys:
                if key in batch:
                    row_batch[key] = batch[key][row:row + 1]

            num_tokens = self.token_counting_func(row_batch)
            if isinstance(num_tokens, dict):
                output['total_tokens'].append(num_tokens['total'])
                output['loss_generating_tokens'].append(
                    num_tokens['loss_generating'],
                )
            else:
                output['total_tokens'].append(num_tokens)
                output['loss_generating_tokens'].append(num_tokens)

        batch['total_tokens'] = output['total_tokens']
        batch['loss_generating_tokens'] = output['loss_generating_tokens']

        return batch


def _validate_cfg(
    dataset_cfg: dict[str, Any],
    tokenizer: PreTrainedTokenizerBase,
):
    eos_token_id = dataset_cfg.get('eos_token_id', None)
    bos_token_id = dataset_cfg.get('bos_token_id', None)

    tokenizer_eos_token_id = getattr(tokenizer, 'eos_token_id', None)
    if eos_token_id is not None and eos_token_id != tokenizer_eos_token_id:
        eos_mismatch_str = f'Provided {eos_token_id=} does not match the eos_token_id of the tokenizer={tokenizer_eos_token_id}.'
        if dataset_cfg.pop('override_eos_token_id_mismatch_error', False):
            log.warning(eos_mismatch_str)
        else:
            raise ValueError(
                eos_mismatch_str +
                ' To override this error, set the override_eos_token_id_mismatch_error flag to True in the dataset config section of the YAML.',
            )

    tokenizer_bos_token_id = getattr(tokenizer, 'bos_token_id', None)
    if bos_token_id is not None and bos_token_id != tokenizer_bos_token_id:
        bos_mismatch_str = f'Provided {bos_token_id=} does not match the bos_token_id of the tokenizer={tokenizer_bos_token_id}.'
        if dataset_cfg.pop('override_bos_token_id_mismatch_error', False):
            log.warning(bos_mismatch_str)
        else:
            raise ValueError(
                bos_mismatch_str +
                ' To override this error, set the override_bos_token_id_mismatch_error flag to True in the dataset config section of the YAML.',
            )

    max_seq_len = dataset_cfg.get('max_seq_len')
    if max_seq_len is not None:
        if max_seq_len != int(max_seq_len):
            raise ValueError('max_seq_len must be an integer')
        dataset_cfg['max_seq_len'] = int(max_seq_len)


def validate_ds_replication(
    dataset_cfg: dict[str, Any],
    tokenizer: PreTrainedTokenizerBase,
    device_batch_size: Union[int, float],
) -> tuple[int, int]:
    _validate_cfg(dataset_cfg, tokenizer)
    if (dataset_cfg.get('seq_parallel_replication', 1) or 1) > 1:
        raise NotImplementedError('Sequence parallelism is not supported.')
    if not isinstance(device_batch_size, int):
        raise ValueError('device_batch_size should be integer.')
    return dataset_cfg.get('replication', 1) or 1, device_batch_size


def get_data_spec(
    dl: Union[Iterable, TorchDataloader],
    dataset_cfg: dict[str, Any],
) -> DataSpec:
    del dataset_cfg
    token_counting_func = get_tokens_per_batch_func()

    return DataSpec(
        dataloader=dl,
        get_num_tokens_in_batch=token_counting_func,
    )


def get_tokens_per_batch_func(
    decoder_only: bool = True,
) -> Callable[[Batch], Union[int, dict[str, int]]]:
    """Returns a callable that counts the number of tokens in a batch.

    Args:
        pad_token_id (int): The id of the padding token.
        decoder_only (bool, optional): Whether to expect the batch to just contain ``input_ids`` (decoder only)
            or to also contain ``decoder_input_ids`` (encoder decoder). Defaults to ``True``.

    Returns:
        Callable[[Batch], int]: A callable that counts the number of tokens in a batch.
    """

    def get_num_tokens_in_batch(batch: Batch) -> Union[int, dict[str, int]]:
        if not isinstance(batch, Mapping) or (
            'attention_mask' not in batch and 'input_ids' not in batch
        ):
            raise ValueError(
                'get_tokens_per_batch_func() requires a batch with an attention_mask key or an input_ids key',
            )

        if not decoder_only and 'decoder_attention_mask' not in batch:
            raise ValueError(
                'get_tokens_per_batch_func() for encoder decoder requires a batch with a decoder_attention_mask key',
            )

        # Short cut if the dataloader has already calculated the number of tokens
        if 'total_tokens' in batch and 'loss_generating_tokens' in batch:
            return {
                'total': sum(batch['total_tokens']),
                'loss_generating': sum(batch['loss_generating_tokens']),
            }

        # Count number of non padding tokens in batch
        if 'attention_mask' in batch:
            input_ids_tokens = int(torch.sum(batch['attention_mask']).item())
        else:
            input_ids_tokens = batch['input_ids'].numel()

        loss_generating_tokens = None
        if 'labels' in batch:
            loss_generating_tokens = (
                batch['labels'].shape[0] * (batch['labels'].shape[1] - 1)
            ) - torch.count_nonzero(
                torch.eq(batch['labels'][..., 1:], CROSS_ENTROPY_IGNORE_INDEX),
            )

        # For encoder decoder models only
        decoder_input_ids_tokens = 0
        if not decoder_only:
            decoder_input_ids_tokens = int(
                torch.sum(batch['decoder_attention_mask']).item(),
            )

        if loss_generating_tokens is not None:
            return {
                'total': input_ids_tokens + decoder_input_ids_tokens,
                'loss_generating': loss_generating_tokens,
            }
        return input_ids_tokens + decoder_input_ids_tokens

    return get_num_tokens_in_batch


def get_text_collator(
    dataloader_cfg: dict[str, Any],
    tokenizer: PreTrainedTokenizerBase,
    dataset_batch_size: int,
) -> tuple[Union[transformers.DataCollatorForLanguageModeling,
                 ConcatenatedSequenceCollatorWrapper,
                 LossGeneratingTokensCollatorWrapper], int]:
    dataset_cfg = dataloader_cfg.get('dataset')
    assert isinstance(dataset_cfg, dict)
    eos_token_id = dataset_cfg.get('eos_token_id', None)
    bos_token_id = dataset_cfg.get('bos_token_id', None)
    mlm_probability = dataset_cfg.pop('mlm_probability', None)
    collate_fn = transformers.DataCollatorForLanguageModeling(
        tokenizer=tokenizer,
        mlm=mlm_probability is not None,
        mlm_probability=mlm_probability,
    )

    if (eos_token_id is not None) or (bos_token_id is not None):
        # Note: Will raise an error if both are non-None
        collate_fn = ConcatenatedSequenceCollatorWrapper(
            base_collator=collate_fn,
            eos_token_id=eos_token_id,
            bos_token_id=bos_token_id,
        )

    collate_fn = LossGeneratingTokensCollatorWrapper(
        collate_fn,
        get_tokens_per_batch_func(),
    )

    return collate_fn, dataset_batch_size


def get_finetuning_collator(
    dataloader_cfg: dict[str, Any],
    tokenizer: PreTrainedTokenizerBase,
    dataset_batch_size: int,
) -> tuple[Union[Seq2SeqFinetuningCollator, BinPackCollator,
                 LossGeneratingTokensCollatorWrapper], int]:
    collate_fn, dataset_batch_size = build_collate_fn(
        dataloader_cfg,
        tokenizer,
        dataset_batch_size,
    )
    collate_fn = LossGeneratingTokensCollatorWrapper(
        collate_fn,
        get_tokens_per_batch_func(),
    )
    return collate_fn, dataset_batch_size



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import os
import platform
import socket
from typing import Any, Optional

import git
import pkg_resources
import psutil
import torch
from composer.core import Callback, State
from composer.loggers import Logger
from composer.utils import dist

from mcli import sdk

__all__ = ['EnvironmentLoggingCallback']

_PACKAGES_TO_LOG = [
    'llm-foundry',
    'mosaicml',
    'megablocks',
    'grouped-gemm',
    'torch',
    'flash_attn',
    'transformers',
    'datasets',
    'peft',
]


class EnvironmentLoggingCallback(Callback):
    """A callback for logging environment information during model training.

    This callback collects various pieces of information about the training environment,
    including git repository details, package versions, system information, GPU details,
    distributed training setup, NVIDIA driver information, and Docker container details.

    Args:
        workspace_dir (str): The directory containing the workspace. Defaults to '/workspace'.
        log_git (bool): Whether to log git repository information. Defaults to True.
        log_packages (bool): Whether to log package versions. Defaults to True.
        log_nvidia (bool): Whether to log NVIDIA driver information. Defaults to True.
        log_docker (bool): Whether to log Docker container information. Defaults to True.
        log_system (bool): Whether to log system information. Defaults to False.
        log_gpu (bool): Whether to log GPU information. Defaults to False.
        log_distributed (bool): Whether to log distributed training information. Defaults to False.
        packages_to_log (list[str]): A list of package names to log versions for. Defaults to None.

    The collected information is logged as hyperparameters at the start of model fitting.
    """

    def __init__(
        self,
        workspace_dir: str = '/workspace',
        log_git: bool = True,
        log_nvidia: bool = True,
        log_docker: bool = True,
        log_packages: bool = True,
        log_system: bool = False,
        log_gpu: bool = False,
        log_distributed: bool = False,
        packages_to_log: Optional[list[str]] = None,
    ):
        self.workspace_dir = workspace_dir
        self.log_git = log_git
        self.log_packages = log_packages
        self.log_nvidia = log_nvidia
        self.log_docker = log_docker
        self.log_system = log_system
        self.log_gpu = log_gpu
        self.log_distributed = log_distributed
        self.env_data: dict[str, Any] = {}
        self.packages_to_log = packages_to_log or _PACKAGES_TO_LOG

    def _get_git_info(self, repo_path: str) -> Optional[dict[str, str]]:
        if not os.path.isdir(repo_path):
            return None
        try:
            repo = git.Repo(repo_path)
            return {
                'commit_hash': repo.head.commit.hexsha,
                'branch': repo.active_branch.name,
            }
        except (git.InvalidGitRepositoryError, git.NoSuchPathError):
            return None

    def _get_package_version(self, package_name: str) -> Optional[str]:
        try:
            return pkg_resources.get_distribution(package_name).version
        except pkg_resources.DistributionNotFound:
            return None

    def _get_system_info(self) -> dict[str, Any]:
        return {
            'python_version': platform.python_version(),
            'os': f'{platform.system()} {platform.release()}',
            'hostname': socket.gethostname(),
            'cpu_info': {
                'model': platform.processor(),
                'cores': psutil.cpu_count(logical=False),
                'threads': psutil.cpu_count(logical=True),
            },
            'memory': {
                'total': psutil.virtual_memory().total,
                'available': psutil.virtual_memory().available,
            },
        }

    def _get_gpu_info(self) -> dict[str, Any]:
        if torch.cuda.is_available():
            return {
                'model': torch.cuda.get_device_name(0),
                'count': torch.cuda.device_count(),
                'memory': {
                    'total': torch.cuda.get_device_properties(0).total_memory,
                    'allocated': torch.cuda.memory_allocated(0),
                },
            }
        return {'available': False}

    def _get_nvidia_info(self) -> dict[str, Any]:
        if torch.cuda.is_available():
            nccl_version = torch.cuda.nccl.version()  # type: ignore
            return {
                'cuda_version':
                    torch.version.cuda,  # type: ignore[attr-defined]
                'cudnn_version': str(
                    torch.backends.cudnn.version(),
                ),  # type: ignore[attr-defined]
                'nccl_version': '.'.join(map(str, nccl_version)),
            }
        return {'available': False}

    def _get_distributed_info(self) -> dict[str, Any]:
        return {
            'world_size': dist.get_world_size(),
            'local_world_size': dist.get_local_world_size(),
            'rank': dist.get_global_rank(),
            'local_rank': dist.get_local_rank(),
        }

    def _get_docker_info(self) -> Optional[dict[str, Any]]:
        if 'RUN_NAME' not in os.environ:
            return None
        run = sdk.get_run(os.environ['RUN_NAME'])
        image, tag = run.image.split(':')
        return {
            'image': image,
            'tag': tag,
        }

    def fit_start(self, state: State, logger: Logger) -> None:
        # Collect environment data
        if self.log_git:
            self.env_data['git_info'] = {}
            for folder in os.listdir(self.workspace_dir):
                path = self._get_git_info(
                    os.path.join(self.workspace_dir, folder),
                )
                if path:
                    self.env_data['git_info'][folder] = path

        if self.log_packages:
            self.env_data['package_versions'] = {
                pkg: self._get_package_version(pkg)
                for pkg in self.packages_to_log
            }
        if self.log_nvidia:
            self.env_data['nvidia'] = self._get_nvidia_info()

        if self.log_docker:
            if docker_info := self._get_docker_info():
                self.env_data['docker'] = docker_info

        if self.log_system:
            self.env_data['system_info'] = self._get_system_info()

        if self.log_gpu:
            self.env_data['gpu_info'] = self._get_gpu_info()

        if self.log_distributed:
            self.env_data['distributed_info'] = self._get_distributed_info()

        # Log the collected data
        logger.log_hyperparameters({'environment_data': self.env_data})



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""LLM Contrastive Embedding Model.

Implements InfoNCE Loss using the MPT architecture. The resulting model can
be used as a vector embedding model.

This is inspired by Microsoft Research's unilm repository
https://github.com/microsoft/unilm
"""

import logging
from dataclasses import dataclass
from typing import Any, Mapping, MutableMapping, Optional, Union, cast

import torch
import torch.distributed
import torch.nn as nn
import torch.nn.functional as F
from composer.models import HuggingFaceModel
from composer.utils import dist
from einops import rearrange
from omegaconf import OmegaConf as om
from torch.distributed.nn.functional import all_gather
from torchmetrics import Metric
from transformers import PreTrainedTokenizerBase
from transformers.modeling_outputs import CausalLMOutputWithPast

from llmfoundry import registry
from llmfoundry.models.hf.hf_causal_lm import ComposerHFCausalLM
from llmfoundry.models.mpt.configuration_mpt import MPTConfig
from llmfoundry.models.mpt.modeling_mpt import MPTForCausalLM
from llmfoundry.models.utils.config_moe_args import create_set_process_group

log = logging.getLogger(__name__)


class ContrastiveEvalLoss(Metric):

    def __init__(self):
        super().__init__()
        self.add_state('loss', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('total', default=torch.tensor(0), dist_reduce_fx='sum')

    def update(self, outputs: Any, labels: Any):
        loss = outputs['loss']
        if loss.device != self.loss.device:
            loss = loss.to(self.loss.device)
        self.loss += loss
        self.total += 1

    def compute(self):
        return self.loss / self.total


@dataclass
class ContrastiveConfig:
    """Configuration for the contrastive loss.

    Args:
        temperature (Union[int, float], optional): Temperature for InfoNCE Loss. Defaults to 1.
        vector_representation (str, optional): The vector representation to use. Defaults to 'avg'.
        normalize_output (bool, optional): Whether to normalize the output. Defaults to True.
        gather_in_batch_negatives (bool, optional): Whether to call all_gather on all samples in global batch
        use_legacy_gradient_passthrough (bool, optional): Whether to use the legacy gradient passthrough. Defaults to False.
        infonce_process_group_size (int, optional): The size of the process group for InfoNCE loss. Defaults to None.
    """
    temperature: Union[int, float] = 1
    vector_representation: str = 'avg'
    normalize_output: bool = True
    pos_step_size: int = -1  #keep for backwards compatibility
    gather_in_batch_negatives: bool = False
    use_legacy_gradient_passthrough: bool = False
    infonce_process_group_size: Optional[int] = None


class ContrastiveModel(HuggingFaceModel):
    """A contrastive loss function wrapping MPT or Huggingface architecture.

    This model applies a contrastive loss function to either a MPT (Mosaic Pretrained Transformer)
    or a Huggingface architecture. It allows for bidirectional encoding by modifying the attention mask.

    Args:
        tokenizer (PreTrainedTokenizerBase): The tokenizer used for tokenization.
        contrastive_config (Dict[str, Any], optional): Configuration for the contrastive loss. Defaults to None. See `ContrastiveConfig`.
        pretrained_model_name_or_path (Optional[str], optional): Pretrained model name or path. Defaults to None.
        pretrained_lora_id_or_path (Optional[str], optional): Pretrained LoRA (Low Rank Adaptation) ID or path. Defaults to None.
        trust_remote_code (bool, optional): Whether to trust remote code. Defaults to False.
        init_device (str, optional): The initial device. Defaults to 'cpu'.
        use_flash_attention_2 (bool, optional): Whether to use Flash Attention 2. Defaults to True.
        use_auth_token (bool, optional): Whether to use an authentication token. Defaults to False.
        config_overrides (Optional[Dict[str, Any]], optional): Overrides for the model configuration. Defaults to None.
        load_in_8bit (bool, optional): Whether to load the model in 8-bit mode. Defaults to False.
        loss_fn (str, optional): The loss function to use (either 'torch_crossentropy' or 'fused_crossentropy'). Defaults to 'fused_crossentropy'.
        pretrained (bool, optional): Whether to use a pretrained model when using a Hugging Face architecture. Defaults to True.
        **kwargs (Dict[str, Any]): Additional keyword arguments.
    """

    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        contrastive_config: Optional[dict[str, Any]] = None,
        pretrained_model_name_or_path: Optional[str] = None,
        pretrained_lora_id_or_path: Optional[str] = None,
        trust_remote_code: bool = False,
        init_device: str = 'cpu',
        use_flash_attention_2: bool = True,
        use_auth_token: bool = False,
        config_overrides: Optional[dict[str, Any]] = None,
        load_in_8bit: bool = False,
        loss_fn: str = 'fused_crossentropy',
        pretrained: bool = True,
        **kwargs: dict[str, Any],
    ):
        self.pretrained_model_name_or_path = pretrained_model_name_or_path
        self.pretrained = pretrained
        self.pretrained_lora_id_or_path = pretrained_lora_id_or_path
        self.trust_remote_code = trust_remote_code
        self.init_device = init_device
        self.use_flash_attention_2 = use_flash_attention_2
        self.use_auth_token = use_auth_token
        self.config_overrides = config_overrides
        self.load_in_8bit = load_in_8bit
        self.kwargs = kwargs
        self.is_mpt = False

        contrastive_config = contrastive_config or {}
        contrastive_config_obj: ContrastiveConfig = om.structured(
            ContrastiveConfig(**contrastive_config),
        )
        if tokenizer.pad_token is None:  # type: ignore
            tokenizer.pad_token = tokenizer.eos_token

        model = self.construct_model()

        train_metrics: list[Metric] = [
        ]  # TODO: no train metrics for embedding models yet!

        self.eval_metrics = [
            ContrastiveEvalLoss(),
        ]

        super().__init__(
            model=model,
            tokenizer=tokenizer,
            use_logits=False,
            metrics=train_metrics,
            eval_metrics=self.eval_metrics, # type: ignore
            shift_labels=False,
            allow_embedding_resizing=True,
        )

        # Temperature for InfoNCELoss
        self.temperature = contrastive_config_obj.temperature

        # Set the vector representation to either be the average of all the individual token vectors,
        # or the EOS token at the end of the sequence
        self.vector_representation = contrastive_config_obj.vector_representation
        self.normalize_output = contrastive_config_obj.normalize_output

        self.step_size = None
        self.gather_in_batch_negatives = contrastive_config_obj.gather_in_batch_negatives
        self.use_legacy_gradient_passthrough = contrastive_config_obj.use_legacy_gradient_passthrough
        self.n_active_params = sum(p.numel() for p in self.parameters())
        if loss_fn == 'fused_crossentropy':
            try:
                from flash_attn.losses.cross_entropy import \
                    CrossEntropyLoss as FusedCrossEntropyLoss

                self.loss_fn = FusedCrossEntropyLoss(ignore_index=-100)
            except:
                raise ValueError(
                    'Fused Cross Entropy is not installed. Either (1) have a CUDA-compatible GPU '
                    +
                    'and `pip install .[gpu]` if installing from source or `pip install xentropy-cuda-lib@git+https://github.com/HazyResearch/flash-attention.git@v1.0.3#subdirectory=csrc/xentropy` '
                    +
                    'if installing from pypi, or (2) set your config model.loss_fn=torch_crossentropy.',
                )
        elif loss_fn == 'torch_crossentropy':
            self.loss_fn = nn.CrossEntropyLoss(ignore_index=-100)
        else:
            raise ValueError(
                f'Specified loss_fn={loss_fn} not recognized. `loss_fn` must be one of [`fused_crossentropy`, `torch_crossentropy`].',
            )

        self.infonce_process_group = None
        if contrastive_config_obj.infonce_process_group_size is not None:
            pg_size = contrastive_config_obj.infonce_process_group_size
            self.infonce_process_group = create_set_process_group(pg_size)

    def construct_model(self):
        if self.pretrained_model_name_or_path:
            model_class = registry.models.get('hf_causal_lm')
            model_class = cast(type[ComposerHFCausalLM], model_class)
            model = model_class.build_inner_model(
                pretrained=self.pretrained,
                pretrained_model_name_or_path=self.
                pretrained_model_name_or_path,
                pretrained_lora_id_or_path=self.pretrained_lora_id_or_path,
                trust_remote_code=self.trust_remote_code,
                init_device=self.init_device,
                use_flash_attention_2=self.use_flash_attention_2,
                use_auth_token=self.use_auth_token,
                config_overrides=self.config_overrides or {},
                load_in_8bit=self.load_in_8bit,
                **self.kwargs,
            )
        else:
            model = MPTForCausalLM(MPTConfig(**self.kwargs))
            self.is_mpt = True
        return model

    def _update_step_size_if_needed(self, batch: MutableMapping) -> None:
        """Update step size on first batch if we detect hard negatives."""
        if self.step_size:
            return

        input_shape = batch['input_ids'].shape
        if input_shape[1] > 2:
            # We have hard negatives, batch shape is [batch, sample of query+positive passage+negative passages, tokens].
            self.step_size = input_shape[1]
            log.info(
                f'Detected hard negatives, updated step_size to {self.step_size}',
            )
        else:
            self.step_size = 2

    def format_queries_batch(
        self,
        batch: MutableMapping,
        last_hidden_state: torch.Tensor,
    ) -> tuple[dict[str, torch.Tensor], torch.Tensor]:
        """Format `queries` by selecting every ``n``th entry from the batch.

        Here ``n`` is the step size, which represents the number of hard
        negatives per passage.
        """
        assert self.step_size
        queries = {}
        indices = list(range(0, batch['input_ids'].size(0), self.step_size))
        for key in batch:
            queries[key] = batch[key][indices, :]
        return queries, last_hidden_state[indices, :, :]

    def format_passages_batch(
        self,
        batch: MutableMapping,
        last_hidden_state: torch.Tensor,
    ) -> tuple[dict[str, torch.Tensor], torch.Tensor]:
        """Format `passages` by selecting every ``n``th entry from the batch.

        Here ``n`` is the step size, which represents the number of hard
        negatives per passage.
        """
        assert self.step_size
        passages = {}
        num_blocks = batch['input_ids'].size(0) // self.step_size
        index = torch.arange(
            1,
            num_blocks * self.step_size + 1,
            device=last_hidden_state.device,
        ).view(num_blocks, self.step_size)
        index = index[:, :self.step_size - 1].reshape(-1)
        for key in batch:
            passages[key] = batch[key][index]
        return passages, last_hidden_state[index, :, :]

    def forward(self, batch: MutableMapping) -> CausalLMOutputWithPast:
        # Collapse pairs into the batch dimension
        self._update_step_size_if_needed(batch)
        collapse_dims = lambda x: rearrange(x, 'b p d -> (b p) d') if \
            len(x.shape) > 2 else x

        for key in batch:
            batch[key] = collapse_dims(batch[key])

        return self.model(
            output_hidden_states=True,
            **batch,
        )

    def _cat_gather(self, t: torch.Tensor, group: Any = None) -> torch.Tensor:
        """Applies an all gather operation necessary for InfoNCELoss.

        See https://github.com/pytorch/pytorch/blob/63d5e9221bedd1546b7d364b5ce4171547db12a9/torch/distributed/nn/functional.py#L314
        as well as https://github.com/pytorch/pytorch/issues/121587#issuecomment-1989070351
        """
        if self.use_legacy_gradient_passthrough:
            all_tensors = list(dist.all_gather(t, group))
            all_tensors[dist.get_global_rank()] = t
            all_tensors = torch.cat(all_tensors)
        else:
            extra_kwargs = {'group': group} if group is not None else {}
            all_tensors = all_gather(t, **extra_kwargs)
            all_tensors = torch.cat(all_tensors)

        return all_tensors

    def get_hidden_state(self, outputs: CausalLMOutputWithPast) -> torch.Tensor:
        """Returns the hidden state to use for pooling."""
        return outputs.hidden_states[-1]

    def handle_language_head(
        self,
        outputs: CausalLMOutputWithPast,
    ) -> torch.Tensor:
        """Handles `zero` tensor to avoid DDP unused parameters error."""
        return torch.sum(
            outputs.logits,
        ) * 0  # This attaches the language head to the computation graph

    def _compute_scores(
        self,
        batch: MutableMapping,
        outputs: CausalLMOutputWithPast,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Run pairs through the encoder separately in two passes.

        This function splits queries and passages based on the step size, which represents
        the number of hard negatives per passage. It then runs the queries and passages
        through the encoder separately to obtain the encoded representations. The encoded
        representations are used for further computations in the model.

        Args:
            batch (MutableMapping): The input batch containing queries and passages.
            outputs (CausalLMOutputWithPast): The model outputs.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: The encoded representations of queries and passages.
        """
        hidden_state = self.get_hidden_state(outputs)
        zero = self.handle_language_head(outputs)
        (
            queries_batch,
            queries_last_hidden_state,
        ) = self.format_queries_batch(batch, hidden_state)
        (
            passages_batch,
            passages_last_hidden_state,
        ) = self.format_passages_batch(batch, hidden_state)

        query_attn_mask = queries_batch.get('attention_mask')
        passage_attn_mask = passages_batch.get('attention_mask')
        assert isinstance(query_attn_mask, torch.Tensor)
        assert isinstance(passage_attn_mask, torch.Tensor)
        if self.vector_representation == 'eos':

            def pool_fn(x: torch.Tensor, mask: torch.Tensor):
                row_indices = torch.arange(mask.shape[0])
                flipped_mask = ~mask.bool()
                last_true_indices = flipped_mask.int().argmax(dim=1) - 1
                pooled_outputs = x[row_indices, last_true_indices, :]
                return pooled_outputs
        elif self.vector_representation == 'avg':

            def pool_fn(x: torch.Tensor, mask: torch.Tensor):
                x = x.masked_fill(~mask[..., None].bool(), 0.0)
                pooled_outputs = x.sum(dim=1) / (mask.sum(dim=1)[..., None])
                return pooled_outputs
        else:
            raise ValueError(
                f'Specified vector_representation={self.vector_representation} not recognized. `vector_representation` must be one of [`avg`, `eos`].',
            )

        q_pooled_outputs = pool_fn(queries_last_hidden_state, query_attn_mask)
        p_pooled_outputs = pool_fn(
            passages_last_hidden_state,
            passage_attn_mask,
        )

        if self.normalize_output:
            q_pooled_outputs = F.normalize(q_pooled_outputs, dim=-1)
            p_pooled_outputs = F.normalize(p_pooled_outputs, dim=-1)

        # Use all_gather to include negatives across mini batch
        if self.gather_in_batch_negatives:
            all_q_pooled_outputs = self._cat_gather(
                q_pooled_outputs,
                group=self.infonce_process_group,
            )
            all_p_pooled_outputs = self._cat_gather(
                p_pooled_outputs,
                group=self.infonce_process_group,
            )
        else:
            all_q_pooled_outputs = q_pooled_outputs
            all_p_pooled_outputs = p_pooled_outputs

        assert all_q_pooled_outputs is not None
        assert all_p_pooled_outputs is not None

        all_scores = self._full_contrastive_scores(
            queries=all_q_pooled_outputs,
            passages=all_p_pooled_outputs,
        )
        all_scores = all_scores * (1 / self.temperature) + zero

        all_labels = torch.arange(
            all_scores.size(0),
            device=q_pooled_outputs.device,
            dtype=torch.long,
        )
        all_labels = all_labels * (
            p_pooled_outputs.size(0) // q_pooled_outputs.size(0)
        )

        return all_scores, all_labels

    def _full_contrastive_scores(
        self,
        queries: torch.Tensor,
        passages: torch.Tensor,
    ) -> torch.Tensor:

        # this calculates the inner product between query and passage pairs
        qp = torch.mm(queries, passages.t())

        return qp

    def loss(
        self,
        outputs: CausalLMOutputWithPast,
        batch: MutableMapping,
    ) -> torch.Tensor:
        scores, labels = self._compute_scores(batch, outputs)
        loss = self.loss_fn(scores, labels)
        return loss

    def eval_forward(
        self,
        batch: MutableMapping,
        outputs: Optional[Any] = None,
    ):
        if outputs is None:
            outputs = self.forward(batch)
        val_loss = self.loss(outputs, batch)
        return {'loss': val_loss, 'outputs': outputs}

    def flops_per_batch(self, batch: Mapping) -> int:
        # Note: this computation does not take into account padding, and assumes
        # that the dataset has been constructed without padding. Additionally, we
        # assume the backward pass is approximately 2x the forward pass

        bs, msl = batch['input_ids'].shape[0:2]
        params_flops_per_token = 2 * self.n_active_params
        params_flops_per_seq = params_flops_per_token * msl
        attn_flops_per_seq = (
            self.model.config.n_layers * 2 * 2 *
            (self.model.config.d_model * (msl**2))
        )

        return (params_flops_per_seq + attn_flops_per_seq) * 3 * bs



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import time

import requests
from omegaconf import DictConfig
from transformers import AutoTokenizer

from llmfoundry.models.inference_api_wrapper.openai_causal_lm import (
    OpenAICausalLMEvalWrapper,
    OpenAIChatAPIEvalWrapper,
    OpenAIEvalInterface,
)

__all__ = [
    'FMAPICasualLMEvalWrapper',
    'FMAPIChatAPIEvalWrapper',
    'FMAPIEvalInterface',
]

log = logging.getLogger(__name__)


class FMAPIEvalInterface(OpenAIEvalInterface):

    def block_until_ready(self, base_url: str):
        """Block until the endpoint is ready."""
        sleep_s = 5
        timeout_s = 5 * 60  # At max, wait 5 minutes

        ping_url = f'{base_url}/ping'

        waited_s = 0
        while True:
            try:
                requests.get(ping_url)
                log.info(f'Endpoint {ping_url} is ready')
                break
            except requests.exceptions.ConnectionError:
                log.debug(
                    f'Endpoint {ping_url} not ready yet. Sleeping {sleep_s} seconds',
                )
                time.sleep(sleep_s)
                waited_s += sleep_s

            if waited_s >= timeout_s:
                raise TimeoutError(
                    f'Endpoint {ping_url} did not become read after {waited_s:,} seconds, exiting',
                )

    def __init__(self, om_model_config: DictConfig, tokenizer: AutoTokenizer):
        is_local = om_model_config.pop('local', False)
        if is_local:
            base_url = os.environ.get(
                'MOSAICML_MODEL_ENDPOINT',
                'http://0.0.0.0:8080/v2',
            )
            om_model_config['base_url'] = base_url
            self.block_until_ready(base_url)

        if 'base_url' not in om_model_config:
            raise ValueError(
                'Must specify base_url or use local=True in model_cfg for FMAPIsEvalWrapper',
            )

        super().__init__(om_model_config, tokenizer)


class FMAPICasualLMEvalWrapper(FMAPIEvalInterface, OpenAICausalLMEvalWrapper):
    """Databricks Foundational Model API wrapper for causal LM models."""


class FMAPIChatAPIEvalWrapper(FMAPIEvalInterface, OpenAIChatAPIEvalWrapper):
    """Databricks Foundational Model API wrapper for chat models."""



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""A collection of common torchmetrics."""

from llmfoundry.eval.metrics.nlp import (
    InContextLearningGenerationExactMatchAccuracy,
    InContextLearningLMAccuracy,
    InContextLearningLMExpectedCalibrationError,
    InContextLearningMCExpectedCalibrationError,
    InContextLearningMetric,
    InContextLearningMultipleChoiceAccuracy,
)

__all__ = [
    'InContextLearningMetric',
    'InContextLearningLMAccuracy',
    'InContextLearningMultipleChoiceAccuracy',
    'InContextLearningGenerationExactMatchAccuracy',
    'InContextLearningLMExpectedCalibrationError',
    'InContextLearningMCExpectedCalibrationError',
]



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Streaming dataset conversion scripts for C4 and The Pile."""
import json
import os
import platform
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Iterable, Optional, Union

import datasets as hf_datasets
import psutil
import torch
from numpy.typing import NDArray
from streaming import MDSWriter
from torch.utils.data import DataLoader, Dataset, IterableDataset
from tqdm import tqdm
from transformers import PreTrainedTokenizerBase

from llmfoundry.data import ConcatTokensDataset, NoConcatDataset
from llmfoundry.utils.builders import build_tokenizer


class ConcatMode(Enum):
    NO_CONCAT = 'NO_CONCAT'
    CONCAT_TOKENS = 'CONCAT_TOKENS'


@dataclass
class DataSplitConstants:
    hf_split: str
    folder_split: str
    raw_samples: Optional[int]
    truncated_samples: Union[int, None]


@dataclass
class DatasetConstants:
    chars_per_sample: int
    chars_per_token: int
    splits: dict[str, DataSplitConstants] = field(default_factory=dict)

    def __iter__(self):
        for v in self.splits.values():
            yield v


class TrainSmallConstants(DataSplitConstants):

    def __init__(
        self,
        hf_split: str = 'train',
        folder_split: str = 'train_small',
        raw_samples: int = 100000,
        truncated_samples: int = 100000,
    ):
        super().__init__(hf_split, folder_split, raw_samples, truncated_samples)


class ValSmallConstants(DataSplitConstants):

    def __init__(
        self,
        hf_split: str = 'validation',
        folder_split: str = 'val_small',
        raw_samples: int = 10000,
        truncated_samples: int = 10000,
    ):
        super().__init__(hf_split, folder_split, raw_samples, truncated_samples)


class ValXSmallConstants(DataSplitConstants):

    def __init__(
        self,
        hf_split: str = 'validation',
        folder_split: str = 'val_xsmall',
        raw_samples: int = 3000,
        truncated_samples: int = 3000,
    ):
        super().__init__(hf_split, folder_split, raw_samples, truncated_samples)


pileconstants = DatasetConstants(
    chars_per_sample=6212,  # Computed over validation set
    chars_per_token=4,  # OpenAI estimate
)
pileconstants.splits['train'] = DataSplitConstants(
    hf_split='train',
    folder_split='train',
    raw_samples=210607728,
    truncated_samples=None,
)
pileconstants.splits['train_small'] = DataSplitConstants(
    hf_split='train',
    folder_split='train_small',
    raw_samples=100000,
    truncated_samples=100000,
)
pileconstants.splits['val'] = DataSplitConstants(
    hf_split='validation',
    folder_split='val',
    raw_samples=214670,
    truncated_samples=None,
)
pileconstants.splits['val_small'] = DataSplitConstants(
    hf_split='validation',
    folder_split='val_small',
    raw_samples=10000,
    truncated_samples=10000,
)
pileconstants.splits['val_xsmall'] = DataSplitConstants(
    hf_split='validation',
    folder_split='val_xsmall',
    raw_samples=3000,
    truncated_samples=3000,
)

c4constants = DatasetConstants(
    chars_per_sample=2163,  # Computed over validation set
    chars_per_token=4,  # OpenAI estimate
)
c4constants.splits['train'] = DataSplitConstants(
    hf_split='train',
    folder_split='train',
    raw_samples=364868892,
    truncated_samples=None,
)
c4constants.splits['train_small'] = DataSplitConstants(
    hf_split='train',
    folder_split='train_small',
    raw_samples=100000,
    truncated_samples=100000,
)
c4constants.splits['val'] = DataSplitConstants(
    hf_split='validation',
    folder_split='val',
    raw_samples=364608,
    truncated_samples=None,
)
c4constants.splits['val_small'] = DataSplitConstants(
    hf_split='validation',
    folder_split='val_small',
    raw_samples=10000,
    truncated_samples=10000,
)
c4constants.splits['val_xsmall'] = DataSplitConstants(
    hf_split='validation',
    folder_split='val_xsmall',
    raw_samples=3000,
    truncated_samples=3000,
)
c4constants.splits['val_xxsmall'] = DataSplitConstants(
    hf_split='validation',
    folder_split='val_xxsmall',
    raw_samples=100,
    truncated_samples=100,
)

CONSTS = {'allenai/c4': c4constants, 'the_pile': pileconstants}


def build_hf_dataset(
    dataset_name: str,
    split: str,
    mode: ConcatMode,
    max_length: Optional[int] = None,
    bos_text: str = '',
    eos_text: str = '',
    no_wrap: bool = False,
    tokenizer: PreTrainedTokenizerBase = None,
    data_subset: Union[str, None] = None,
) -> IterableDataset:
    """Build an IterableDataset over the HF C4 or pile source data.

    Args:
        dataset_name (str): Dataset name
        split (str): Split name.
        mode (ConcatMode): NO_CONCAT, or CONCAT_TOKENS
        max_length (int): The length of concatenated tokens
        bos_text (str): text to insert at the beginning of each sequence
        eos_text (str): text to insert at the end of each sequence
        no_wrap (bool): if concatenating, whether to wrap text across `max_length` boundaries
        tokenizer (PreTrainedTokenizerBase): if mode is CONCAT_TOKENS, the tokenizer to use
        data_subset (str): Referred to as "name" in HuggingFace datasets.load_dataset.
            Typically "all" (The Pile) or "en" (c4).

    Returns:
        An IterableDataset.
    """
    hf_dataset = hf_datasets.load_dataset(
        path=dataset_name,
        name=data_subset,
        split=split,
        streaming=True,
    )
    if mode == ConcatMode.NO_CONCAT:
        dataset = NoConcatDataset(hf_dataset)
    else:
        if not isinstance(tokenizer, PreTrainedTokenizerBase):
            raise ValueError(
                f'{tokenizer=} must be of type PreTrainedTokenizerBase',
            )
        if max_length is None:
            raise ValueError(f'max_length must be set.')
        if bos_text + eos_text == '':
            test_tokens = tokenizer('test')
            if test_tokens['input_ids'][
                0] != tokenizer.bos_token_id and test_tokens['input_ids'][
                    -1] != tokenizer.eos_token_id:
                tok_error_msg = 'This tokenizer does not insert an EOS nor BOS token. '
                tok_error_msg += 'Concatenating with this tokenizer will result in sequences being '
                tok_error_msg += 'attached without a separating token. Please use another tokenizer, '
                tok_error_msg += 'such as facebook/opt-125m, or specify EOS/BOS text with e.g. '
                tok_error_msg += '--bos_text=<|endoftext|>.'
                raise ValueError(tok_error_msg)
        dataset = ConcatTokensDataset(
            hf_dataset=hf_dataset,
            tokenizer=tokenizer,
            max_length=max_length,
            bos_text=bos_text,
            eos_text=eos_text,
            no_wrap=no_wrap,
        )
    return dataset


def _est_progress_denominator(
    total_samples: int,
    chars_per_sample: int,
    chars_per_token: int,
    mode: ConcatMode,
    max_length: int,
):
    est_tokens_per_sample = chars_per_sample // chars_per_token
    if mode == ConcatMode.NO_CONCAT:
        return total_samples
    elif mode == ConcatMode.CONCAT_TOKENS:
        return total_samples * est_tokens_per_sample // max_length


def build_dataloader(
    dataset: Dataset,
    batch_size: int,
    num_workers: Optional[int],
) -> DataLoader:
    if num_workers is None:
        # Multiple workers is only supported on linux machines
        if 'linux' or 'macos' in platform.platform().lower():
            num_workers = max(1, psutil.cpu_count())
        else:
            num_workers = 0

    # If using multiple workers, configure each worker to prefetch as many samples as it can, up to
    # the aggregate device batch size
    # If not using workers, the torch DataLoader expects the default value for prefetch_factor,
    # which non-intuitively must be 2.
    prefetch_factor = max(
        1,
        2 * batch_size // num_workers,
    ) if num_workers > 0 else 2

    return DataLoader(
        dataset=dataset,
        sampler=None,
        batch_size=batch_size,
        num_workers=num_workers,
        prefetch_factor=prefetch_factor,
    )


def generate_samples(
    loader: DataLoader,
    truncate_num_samples: Optional[int] = None,
) -> Iterable[Union[dict[str, bytes], dict[str, NDArray]]]:
    """Generator over samples of a dataloader.

    Args:
       loader (DataLoader): A dataloader emitting batches like {key: [sample0_bytes, sample1_bytes, sample2_bytes, ...]}
       truncate_num_samples (Optional[int]): An optional # of samples to stop at.

    Yields:
        Sample dicts.
    """
    n_samples = 0
    for batch in loader:
        keys = list(batch.keys())
        current_bs = len(batch[keys[0]])
        for idx in range(current_bs):
            if truncate_num_samples is not None and n_samples == truncate_num_samples:
                return
            n_samples += 1
            yield {
                k:
                v[idx].numpy() if isinstance(v[idx], torch.Tensor) else v[idx]
                for k, v in batch.items()
            }


def convert_dataset_hf(
    dataset: str,
    data_subset: Optional[str],
    splits: list[str],
    out_root: str,
    compression: Optional[str],
    concat_tokens: Optional[int],
    tokenizer: Optional[str],
    tokenizer_kwargs: dict[str, Any],
    bos_text: str,
    eos_text: str,
    no_wrap: bool,
    num_workers: Optional[int],
) -> None:
    """Converts HuggingFace datasets to MDS format.

    Args:
        dataset (str): Name of the dataset
        data_subset (Optional[str]): Subset of the dataset (e.g., "all" or "en")
        splits (list[str]): Comma-separated list of dataset splits
        out_root (str): Output root directory
        compression (Optional[str]): Compression type
        concat_tokens (Optional[int]): Concatenate tokens up to this many tokens
        tokenizer (Optional[str]): Tokenizer name
        tokenizer_kwargs (dict[str, Any]): Tokenizer keyword arguments
        bos_text (str): BOS text
        eos_text (str): EOS text
        no_wrap (bool): Do not wrap text across max_length boundaries
        num_workers (Optional[int]): Number of workers

    Raises:
        KeyError: If constants are not defined for the split
    """
    try:
        dataset_constants = CONSTS[dataset]
    except KeyError:
        raise ValueError(
            f'Constants for dataset "{dataset}" not found. Currently only "the_pile" and "allenai/c4" are supported.',
        )

    if concat_tokens is not None and tokenizer is not None:
        mode = ConcatMode.CONCAT_TOKENS
        built_tokenizer = build_tokenizer(tokenizer, tokenizer_kwargs)
        # we will enforce length, so suppress warnings about sequences too long for the model
        built_tokenizer.model_max_length = int(1e30)
        columns = {'tokens': 'ndarray:int32'}
    else:
        mode = ConcatMode.NO_CONCAT
        built_tokenizer = None
        columns = {'text': 'str'}

    for split_name in splits:
        try:
            split = dataset_constants.splits[split_name]
        except KeyError:
            raise KeyError(f'Constants not defined for split {split_name}.')
        hf_split = split.hf_split
        folder_split = split.folder_split
        expected_num_samples = split.raw_samples
        truncate_num_samples = split.truncated_samples
        # Only generate the splits requested
        if folder_split not in splits:
            continue

        # Get samples
        hf_dataset = build_hf_dataset(
            dataset_name=dataset,
            data_subset=data_subset,
            split=hf_split,
            mode=mode,
            max_length=concat_tokens,
            bos_text=bos_text,
            eos_text=eos_text,
            no_wrap=no_wrap,
            tokenizer=built_tokenizer,
        )
        loader = build_dataloader(
            dataset=hf_dataset,
            batch_size=512,
            num_workers=num_workers,
        )
        samples = generate_samples(
            loader,
            truncate_num_samples=truncate_num_samples,
        )

        if expected_num_samples is not None and concat_tokens is not None:
            denominator = truncate_num_samples if truncate_num_samples is not None else _est_progress_denominator(
                total_samples=expected_num_samples,
                chars_per_sample=dataset_constants.chars_per_sample,
                chars_per_token=dataset_constants.chars_per_token,
                mode=mode,
                max_length=concat_tokens,
            )
        else:
            denominator = None

        # Write samples
        print(f'Converting {folder_split} to MDS format...')
        print(
            f'Note: the progress bar is based on the dataset length before tokenization, and may finish at a value before 100%.',
        )
        with MDSWriter(
            columns=columns,
            out=os.path.join(out_root, folder_split),
            compression=compression,
        ) as out:
            if denominator is not None:
                for sample in tqdm(
                    samples,
                    desc=folder_split,
                    total=denominator,
                ):
                    out.write(sample)
            else:
                for sample in tqdm(samples, desc=folder_split):
                    out.write(sample)


def convert_dataset_hf_from_args(
    dataset: str,
    data_subset: Optional[str],
    splits: list[str],
    out_root: str,
    compression: Optional[str],
    concat_tokens: Optional[int],
    tokenizer: Optional[str],
    tokenizer_kwargs: Optional[str],
    bos_text: Optional[str],
    eos_text: Optional[str],
    no_wrap: bool,
    num_workers: Optional[int],
) -> None:
    """A wrapper for `convert_dataset_hf` that parses arguments.

    Args:
        dataset (str): Name of the dataset
        data_subset (Optional[str]): Subset of the dataset (e.g., "all" or "en")
        splits (list[str]): Comma-separated list of dataset splits
        out_root (str): Output root directory
        compression (Optional[str]): Compression type
        concat_tokens (Optional[int]): Concatenate tokens up to this many tokens
        tokenizer (Optional[str]): Tokenizer name
        tokenizer_kwargs (Optional[str]): Tokenizer keyword arguments in JSON format
        bos_text (Optional[str]): BOS text
        eos_text (Optional[str]): EOS text
        no_wrap (bool): Do not wrap text across max_length boundaries
        num_workers (Optional[int]): Number of workers

    Raises:
        ValueError: If the output directory already contains the requested splits
        ValueError: If `concat_tokens` is set but `tokenizer` is not
    """
    os.environ['WORLD_SIZE'] = '1'
    if tokenizer_kwargs:
        parsed_tokenizer_kwargs = json.loads(tokenizer_kwargs)
    else:
        parsed_tokenizer_kwargs = {}

    if os.path.isdir(out_root) and len(
        set(os.listdir(out_root)).intersection(set(splits)),
    ) > 0:
        raise ValueError(
            f'--out_root={out_root} contains {os.listdir(out_root)} which cannot overlap with the requested splits {splits}.',
        )

    # Make sure we have needed concat options
    if (
        concat_tokens is not None and isinstance(concat_tokens, int) and
        tokenizer is None
    ):
        raise ValueError(
            'When setting --concat_tokens, you must specify a --tokenizer',
        )

    # now that we have validated them, change BOS/EOS to strings and convert
    convert_dataset_hf(
        dataset=dataset,
        data_subset=data_subset,
        splits=splits,
        out_root=out_root,
        compression=compression,
        concat_tokens=concat_tokens,
        tokenizer=tokenizer,
        tokenizer_kwargs=parsed_tokenizer_kwargs,
        bos_text=bos_text if bos_text else '',
        eos_text=eos_text if eos_text else '',
        no_wrap=no_wrap,
        num_workers=num_workers,
    )



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Helper functions for computing parameter counts for MPT model.

Use if generic `sum(p.numel() for p in self.parameters())`
style computation does not account for MoE parameter sharding.
The helper functions in this file account for MoE parameter
sharding in the parameter count calculation. The functions below
calculate the total parameter count and the active parameter count.
Note: MPT has both n_total_params and n_active_params methods.
"""

from typing import Union

from torch import Tensor, nn
from torch.distributed._tensor import DTensor

from llmfoundry.layers_registry import ffns_with_megablocks

__all__ = [
    'mpt_get_active_params',
    'mpt_get_total_params',
]


def module_n_params(module: nn.Module) -> int:
    """Gets the number of parameters in this module excluding child modules.

    Args:
        module (nn.Module): Module of which we get the number of parameters.

    Returns:
        An int for the number of parameters in this module.
    """
    n_params = 0
    for p in module.parameters(recurse=False):
        n_params += p.numel()
    return n_params


def _dtensor_safe_check_numel(tensor: Union[Tensor, DTensor]) -> int:
    if isinstance(tensor, DTensor):
        tensor = tensor._local_tensor
    return tensor.numel()


def megablocks_n_total_params(mpt_model) -> int:  # type: ignore
    """Calculates the number of parameters in a MegaBlocks enabled MPT model.

    MoE experts are sharded across workers. This function scans for MegaBlocks
    modules then multiplies expert params count by MoE world size.

    Args:
        mpt_model (ComposerMPTCausalLM): MPT model of which the number of
            parameters is calculated.

    Returns:
        An int for the total number of parameters in this MPT model.
    """
    import megablocks

    moe_world_size = mpt_model.config.ffn_config.get('moe_world_size')

    n_total_params = 0
    for module in mpt_model.modules():
        if isinstance(
            module,
            (megablocks.layers.mlp.SparseMLP, megablocks.layers.mlp.MLP),
        ):
            n_w1 = _dtensor_safe_check_numel(module.w1)
            n_total_params += n_w1 * moe_world_size
            n_w2 = _dtensor_safe_check_numel(module.w2)
            n_total_params += n_w2 * moe_world_size

            # GLU has an extra weight
            if hasattr(module, 'v1'):
                n_v1 = _dtensor_safe_check_numel(module.v1)
                n_total_params += n_v1 * moe_world_size
        else:
            n_total_params += module_n_params(module)

    return n_total_params


def megablocks_n_active_params(mpt_model) -> int:  # type: ignore
    """Calculates the number of active parameters in a MegaBlocks enabled MPT.

    This requires we calculate the number of elements per expert and
    multiply this by top k.

    Args:
        mpt_model (ComposerMPTCausalLM): MPT model of which the number of
            active parameters is calculated.

    Returns:
        An int for the active number of parameters in this MPT model.
    """
    import megablocks

    moe_num_experts = mpt_model.config.ffn_config.get('moe_num_experts', 1)
    moe_world_size = mpt_model.config.ffn_config.get('moe_world_size')

    local_experts = moe_num_experts / moe_world_size  # if local_experts is < 1, then the expert is sharded

    moe_top_k = mpt_model.config.ffn_config.get('moe_top_k', 1)
    n_active_params = 0
    for module in mpt_model.modules():
        if isinstance(
            module,
            (megablocks.layers.mlp.SparseMLP, megablocks.layers.mlp.MLP),
        ):
            n_w1 = _dtensor_safe_check_numel(module.w1)
            n_active_params += int(n_w1 / local_experts * moe_top_k)
            n_w2 = _dtensor_safe_check_numel(module.w2)
            n_active_params += int(n_w2 / local_experts * moe_top_k)

            # GLU has an extra weight
            if hasattr(module, 'v1'):
                n_v1 = _dtensor_safe_check_numel(module.v1)
                n_active_params += int(n_v1 / local_experts * moe_top_k)
        else:
            n_active_params += module_n_params(module)

    return n_active_params


def mpt_get_total_params(mpt_model) -> int:  # type: ignore
    """Calculates the total parameter count of an MPT model.

    Note: Must be called before model parameters are sharded by FSDP.

    Args:
        mpt_model (ComposerMPTCausalLM): MPT model of which the number of
            active parameters is calculated.

    Returns:
        An int for the total number of parameters in this MPT model.
    """
    if mpt_model.config.ffn_config['ffn_type'] in ffns_with_megablocks:
        return megablocks_n_total_params(mpt_model)
    else:
        return sum(p.numel() for p in mpt_model.parameters())


def mpt_get_active_params(mpt_model) -> int:  # type: ignore
    """Calculates the total parameter count of an MPT model.

    Note: Must be called before model parameters are sharded by FSDP.

    Args:
        mpt_model (ComposerMPTCausalLM): MPT model of which the number of
            active parameters is calculated.

    Returns:
        An int for the active number of parameters in this MPT model.
    """
    if mpt_model.config.ffn_config['ffn_type'] in ffns_with_megablocks:
        params = megablocks_n_active_params(mpt_model)
    else:
        params = sum(p.numel() for p in mpt_model.parameters())
    if not mpt_model.model.transformer.config.tie_word_embeddings:
        # Embedding layers are lookup tables, therefore are not counted in the FLOP computation
        params -= _dtensor_safe_check_numel(
            mpt_model.model.transformer.wte.weight,
        )
    return params



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import datetime

__all__ = [
    'get_date_string',
]


def get_date_string() -> str:
    """Get the current date string."""
    return datetime.datetime.now().strftime('%d %b %Y')



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import os
from typing import Annotated, Optional

import psutil
from typer import Option, Typer

from llmfoundry.command_utils import (
    convert_dataset_hf_from_args,
    convert_dataset_json_from_args,
    convert_delta_to_json_from_args,
    convert_finetuning_dataset_from_args,
    convert_text_to_mds_from_args,
)

app = Typer(pretty_exceptions_show_locals=False)


@app.command(name='convert_dataset_hf')
def convert_dataset_hf(
    dataset: Annotated[str, Option(..., help='Name of the dataset')],
    out_root: Annotated[str, Option(..., help='Output root directory')],
    data_subset: Annotated[
        Optional[str],
        Option(help='Subset of the dataset (e.g., "all" or "en")'),
    ] = None,
    splits: Annotated[str,
                      Option(help='Comma-separated list of dataset splits',),
                     ] = 'train, train_small, val, val_small, val_xsmall',
    compression: Annotated[Optional[str],
                           Option(help='Compression type')] = None,
    concat_tokens: Annotated[
        Optional[int],
        Option(help='Concatenate tokens up to this many tokens')] = None,
    tokenizer: Annotated[Optional[str],
                         Option(help='Tokenizer name')] = None,
    tokenizer_kwargs: Annotated[
        Optional[str],
        Option(help='Tokenizer keyword arguments in JSON format')] = None,
    bos_text: Annotated[Optional[str], Option(help='BOS text')] = None,
    eos_text: Annotated[Optional[str], Option(help='EOS text')] = None,
    no_wrap: Annotated[
        bool,
        Option(help='Do not wrap text across max_length boundaries'),
    ] = False,
    num_workers: Annotated[Optional[int],
                           Option(help='Number of workers')] = None,
):
    """Converts dataset from HuggingFace into JSON files."""
    # Convert comma-separated splits into a list
    splits_list = splits.split(',') if splits else []
    convert_dataset_hf_from_args(
        dataset=dataset,
        data_subset=data_subset,
        splits=splits_list,
        out_root=out_root,
        compression=compression,
        concat_tokens=concat_tokens,
        tokenizer=tokenizer,
        tokenizer_kwargs=tokenizer_kwargs,
        bos_text=bos_text,
        eos_text=eos_text,
        no_wrap=no_wrap,
        num_workers=num_workers,
    )


@app.command(name='convert_dataset_json')
def convert_dataset_json(
    path: Annotated[str, Option(..., help='Path to the input data file')],
    out_root: Annotated[str, Option(..., help='Output root directory')],
    concat_tokens: Annotated[
        int,
        Option(
            ...,
            help='Convert text to tokens and concatenate up to this many tokens',
        )],
    tokenizer: Annotated[str, Option(..., help='Tokenizer name')],
    compression: Annotated[Optional[str],
                           Option(help='Compression type, if any')] = 'zstd',
    split: Annotated[str, Option(help='Dataset split to process')] = 'train',
    bos_text: Annotated[
        Optional[str],
        Option(help='Text to insert at the beginning of each sequence')] = None,
    eos_text: Annotated[
        Optional[str],
        Option(help='Text to insert at the end of each sequence')] = None,
    no_wrap: Annotated[
        bool,
        Option(help='Do not wrap text across max_length boundaries')] = False,
    num_workers: Annotated[
        Optional[int],
        Option(help='Number of workers for data loading')] = None,
):
    """Convert a dataset from JSON to MDS streaming format."""
    convert_dataset_json_from_args(
        path=path,
        split=split,
        out_root=out_root,
        compression=compression,
        concat_tokens=concat_tokens,
        tokenizer=tokenizer,
        bos_text=bos_text,
        eos_text=eos_text,
        no_wrap=no_wrap,
        num_workers=num_workers,
    )


@app.command(name='convert_finetuning_dataset')
def convert_finetuning_dataset_cli(
    dataset: Annotated[
        str,
        Option(
            ...,
            help=
            'Name of the dataset (e.g., first argument to `datasets.load_dataset`, for jsonl data format, it is `json`).',
        )],
    data_subset: Annotated[
        Optional[str],
        Option(help='(Optional) subset of data to use.',)] = None,
    splits: Annotated[str,
                      Option(help='Comma-separated list of dataset splits'),
                     ] = 'train,validation',
    preprocessor: Annotated[
        Optional[str],
        Option(
            help=
            'Name or import path of function used to preprocess (reformat) the dataset.',
        )] = None,
    data_files: Annotated[
        str, Option(help='Data file for each split. Comma-separated.')] = '',
    skip_preprocessing: Annotated[
        bool, Option(help='Whether to skip preprocessing.')] = False,
    out_root: Annotated[
        str,
        Option(
            ...,
            help=
            'Root path of output directory where MDS shards will be stored. Can be a remote URI.',
        )] = '',
    local: Annotated[
        Optional[str],
        Option(
            help=
            '(Optional) root path of local directory if you want to keep a local copy when out_root is remote.',
        )] = None,
    compression: Annotated[
        Optional[str],
        Option(help='(Optional) name of compression algorithm to use.')] = None,
    num_workers: Annotated[Optional[int],
                           Option(help='Number of workers.')] = None,
    tokenizer: Annotated[Optional[str],
                         Option(help='Tokenizer used for processing.')] = None,
    tokenizer_kwargs: Annotated[
        Optional[str],
        Option(
            help=
            'Keyword arguments for tokenizer initialization in JSON format.',
        )] = None,
    max_seq_len: Annotated[int, Option(help='Maximum sequence length.')] = 2048,
    target_prompts: Annotated[
        str,
        Option(help='Policy for when to use prompts as training targets.'),
    ] = 'none',
    target_responses: Annotated[
        str,
        Option(help='Policy for which responses to treat as training targets.'),
    ] = 'last',
    encoder_decoder: Annotated[
        bool,
        Option(
            help=
            'Set if the data are intended to be used to train an encoder-decoder model.',
        )] = False,
):
    """Convert a Finetuning Dataset to MDS streaming format."""
    # Convert comma-separated args
    splits_list = splits.split(',') if splits else []
    data_files_list = data_files.split(',') if data_files else []
    convert_finetuning_dataset_from_args(
        dataset=dataset,
        data_subset=data_subset,
        splits=splits_list,
        preprocessor=preprocessor,
        data_files=data_files_list,
        skip_preprocessing=skip_preprocessing,
        out_root=out_root,
        local=local,
        compression=compression,
        num_workers=num_workers,
        tokenizer=tokenizer,
        tokenizer_kwargs=tokenizer_kwargs,
        max_seq_len=max_seq_len,
        target_prompts=target_prompts,
        target_responses=target_responses,
        encoder_decoder=encoder_decoder,
    )


@app.command(name='convert_text_to_mds')
def convert_text_to_mds(
    output_folder: Annotated[str, Option(..., help='The folder to write output to')],
    input_folder: Annotated[str, Option(..., help='The folder with text files to convert to MDS')],
    concat_tokens: Annotated[int, Option(..., help='Convert text to tokens and concatenate up to this many tokens')],
    tokenizer: Annotated[str, Option(..., help='The name of the tokenizer to use')],
    bos_text: Annotated[Optional[str], Option(help='The text to prepend to each example to separate concatenated examples')] = None,
    eos_text: Annotated[Optional[str], Option(help='The text to append to each example to separate concatenated examples')] = None,
    compression: Annotated[str, Option(help='The compression algorithm to use for MDS writing')] = 'zstd',
    use_tokenizer_eos: Annotated[bool, Option(help='Use the EOS text from the tokenizer')] = False,
    no_wrap: Annotated[bool, Option(help='Whether to let text examples wrap across multiple training examples')] = False,
    processes: Annotated[int, Option(
        help='The number of processes to use to download and convert the dataset',
    )] = min(max(psutil.cpu_count() - 2, 1), 32), # type: ignore
    reprocess: Annotated[bool, Option(
        help=
        'If true, reprocess the input_folder to MDS format. Otherwise, only reprocess upon changes to the input folder or dataset creation parameters.',
    )] = False,
    trust_remote_code: Annotated[bool, Option(
        help='If true, allows custom code to be executed to load the tokenizer',
    )] = False,
    logging_level: Annotated[str, Option(
        help='Logging level for the script. Default is INFO.',
    )] = 'INFO',

):
    """Convert text files to MDS streaming format."""
    convert_text_to_mds_from_args(
        output_folder=output_folder,
        input_folder=input_folder,
        compression=compression,
        concat_tokens=concat_tokens,
        tokenizer_name=tokenizer,
        bos_text=bos_text,
        eos_text=eos_text,
        use_tokenizer_eos=use_tokenizer_eos,
        no_wrap=no_wrap,
        processes=processes,
        reprocess=reprocess,
        trust_remote_code=trust_remote_code,
        logging_level=logging_level,
    )


@app.command(name='convert_delta_to_json')
def convert_delta_to_json_cli(
    delta_table_name: Annotated[str, Option(..., help='UC table <catalog>.<schema>.<table name>')],
    json_output_folder: Annotated[str, Option(..., help='Local path to save the converted json')],
    http_path: Annotated[Optional[str], Option(help='If set, dbsql method is used')] = None,
    batch_size: Annotated[int, Option(help='Row chunks to transmit a time to avoid OOM')] = 1 << 30,
    processes: Annotated[int, Option(help='Number of processes allowed to use')] = os.cpu_count(), # type: ignore
    cluster_id: Annotated[Optional[str], Option(help='Cluster ID with runtime newer than 14.1.0 and access mode of either assigned or shared can use databricks-connect.')] = None,
    use_serverless: Annotated[bool, Option(help='Use serverless or not. Make sure the workspace is entitled with serverless')] = False,
    json_output_filename: Annotated[str, Option(help='The name of the combined final jsonl that combines all partitioned jsonl')] = 'train-00000-of-00001.jsonl',
):
    """Convert a Delta table into JSON files."""
    convert_delta_to_json_from_args(
        delta_table_name=delta_table_name,
        json_output_folder=json_output_folder,
        http_path=http_path,
        batch_size=batch_size,
        processes=processes,
        cluster_id=cluster_id,
        use_serverless=use_serverless,
        json_output_filename=json_output_filename,
    )



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import copy
from typing import Any, Union

import catalogue
import pytest
import torch
from omegaconf import DictConfig
from transformers import PreTrainedTokenizerBase

from llmfoundry.registry import config_transforms
from llmfoundry.utils.config_utils import (
    TRAIN_CONFIG_KEYS,
    TrainConfig,
    make_dataclass_and_log_config,
)


def generate_exclusive_test_params(param_names: list[str]):
    """Generates pytest.param objects with one true parameter for testing.

    Creates pytest.param objects for each parameter name given. For each
    param object, one parameter is set to True (indicating a test case for
    malformed data) while the rest are set to False.

    Args:
        param_names (List[str]): The names of parameters to create test cases for.

    Yields:
        pytest.param: Each with one parameter set to True, indicating the specific case being tested.
    """
    for _, name in enumerate(param_names):
        params = {param_name: False for param_name in param_names}
        params[name] = True
        param_values = list(params.values())
        param_id = f'{name}=True'
        yield pytest.param(*param_values, id=param_id)


def test_config_transforms():
    config = DictConfig({
        'global_train_batch_size': 1,
        'device_train_microbatch_size': 1,
        'model': {},
        'scheduler': {},
        'max_seq_len': 128,
        'train_loader': {},
        'max_duration': 1,
        'tokenizer': {},
        'eval_interval': 1,
        'seed': 1,
        'optimizer': {},
        'variables': {},
    },)

    def dummy_transform(config: dict[str, Any]) -> dict[str, Any]:
        config['variables']['fake_key'] = 'fake_value'
        return config

    config_transforms.register('dummy_transform', func=dummy_transform)

    _, parsed_config = make_dataclass_and_log_config(
        config,
        TrainConfig,
        TRAIN_CONFIG_KEYS,
        transforms='all',
    )

    assert isinstance(parsed_config.variables, dict)
    assert parsed_config.variables['fake_key'] == 'fake_value'

    del catalogue.REGISTRY[
        ('llmfoundry', 'config_transforms', 'dummy_transform')]


def test_logged_cfg():
    config = DictConfig({
        'global_train_batch_size': 1,
        'device_train_microbatch_size': 1,
        'model': {},
        'scheduler': {},
        'max_seq_len': 128,
        'train_loader': {},
        'max_duration': 1,
        'tokenizer': {},
        'eval_interval': 1,
        'seed': 1,
        'optimizer': {},
        'variables': {},
    },)
    logged_config, _ = make_dataclass_and_log_config(
        config,
        TrainConfig,
        TRAIN_CONFIG_KEYS,
        transforms='all',
    )
    expected_config = copy.deepcopy(config)
    expected_config.update({
        'n_gpus': 1,
        'device_train_batch_size': 1,
        'device_train_grad_accum': 1,
        'device_eval_batch_size': 1,
    })
    assert expected_config == logged_config


class MockTokenizer(PreTrainedTokenizerBase):

    def __init__(self) -> None:
        super().__init__()
        self.pad_token: str = '<pad>'
        self.eos_token: str = '</s>'
        self.bos_token: str = '<s>'
        self.unk_token: str = '<unk>'
        self._vocab_size: int = 128

    def __len__(self) -> int:
        return self._vocab_size

    def convert_tokens_to_ids(
        self,
        tokens: Union[str, list[str]],
    ) -> Union[int, list[int]]:
        return 0

    @property
    def pad_token_id(self) -> int:
        return 0

    def _batch_encode_plus(self, *args: Any,
                           **kwargs: Any) -> dict[str, torch.Tensor]:
        batch_texts = args[0] if args else kwargs.get(
            'batch_text_or_text_pairs',
            [],
        )
        max_length = kwargs.get('max_length', 1024)

        if isinstance(batch_texts[0], list):
            texts = [t for pair in batch_texts for t in pair]
        else:
            texts = batch_texts

        token_ids = torch.tensor([
            [hash(text) % 1000 + j for j in range(max_length)] for text in texts
        ])

        return {
            'input_ids': token_ids,
            'attention_mask': torch.ones_like(token_ids),
        }



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from composer.optim import (
    ConstantWithWarmupScheduler,
    CosineAnnealingWithWarmupScheduler,
    DecoupledAdamW,
    LinearWithWarmupScheduler,
)

from llmfoundry.optim.adaptive_lion import DecoupledAdaLRLion, DecoupledClipLion
from llmfoundry.optim.lion import DecoupledLionW
from llmfoundry.optim.no_op import NoOp
from llmfoundry.optim.scheduler import InverseSquareRootWithWarmupScheduler
from llmfoundry.registry import optimizers, schedulers

optimizers.register('adalr_lion', func=DecoupledAdaLRLion)
optimizers.register('clip_lion', func=DecoupledClipLion)
optimizers.register('decoupled_lionw', func=DecoupledLionW)
optimizers.register('decoupled_adamw', func=DecoupledAdamW)
optimizers.register('no_op', func=NoOp)

schedulers.register('constant_with_warmup', func=ConstantWithWarmupScheduler)
schedulers.register(
    'cosine_with_warmup',
    func=CosineAnnealingWithWarmupScheduler,
)
schedulers.register('linear_decay_with_warmup', func=LinearWithWarmupScheduler)
schedulers.register(
    'inv_sqrt_with_warmup',
    func=InverseSquareRootWithWarmupScheduler,
)

__all__ = [
    'DecoupledLionW',
    'DecoupledClipLion',
    'DecoupledAdaLRLion',
    'NoOp',
    'InverseSquareRootWithWarmupScheduler',
]



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import time
from typing import Any, Optional, Union

import pandas as pd
import torch
from composer.core import Callback
from composer.loggers.logger_destination import LoggerDestination
from composer.trainer import Trainer
from composer.utils import dist, get_device, parallelism, reproducibility
from omegaconf import DictConfig
from omegaconf import OmegaConf as om

from llmfoundry.utils import (
    find_mosaicml_logger,
    log_eval_analytics,
    maybe_create_mosaicml_logger,
)
from llmfoundry.utils.builders import (
    add_metrics_to_eval_loaders,
    build_callback,
    build_composer_model,
    build_evaluators,
    build_logger,
    build_tokenizer,
)
from llmfoundry.utils.config_utils import (
    EVAL_CONFIG_KEYS,
    EvalConfig,
    log_config,
    make_dataclass_and_log_config,
    process_init_device,
)
from llmfoundry.utils.registry_utils import import_file

log = logging.getLogger(__name__)


def evaluate_model(
    tokenizer: dict[str, Any],
    model_name: str,
    model: dict[str, Any],
    dist_timeout: Union[float, int],
    run_name: str,
    seed: int,
    icl_tasks: Union[str, list[dict[str, Any]]],
    max_seq_len: int,
    device_eval_batch_size: Union[int, float],
    eval_gauntlet_config: Optional[Union[str, dict[str, Any]]],
    eval_loader_config: Optional[Union[dict[str, Any], list[dict[str, Any]]]],
    loggers: list[LoggerDestination],
    python_log_level: Optional[str],
    precision: str,
    eval_gauntlet_df: Optional[pd.DataFrame],
    eval_subset_num_batches: int,
    icl_subset_num_batches: Optional[int],
    callback_configs: Optional[dict[str, Any]],
    metadata: Optional[dict[str, str]],
    logged_config: dict[str, Any],
    parallelism_config: Optional[dict[str, Any]] = None,
    should_log_config: bool = True,
    load_path: Optional[str] = None,
):
    if parallelism_config:
        deprecated_fsdp_args = list(
            parallelism.FSDPConfig.__annotations__.keys(),
        )
        for deprecated_arg in deprecated_fsdp_args:
            if deprecated_arg in parallelism_config:
                raise ValueError(
                    'parallelism_config cannot contain deprecated fsdp_config arguments.',
                )

    log.info(f'Evaluating model: {model_name}')
    # Build tokenizer and model
    tokenizer_cfg = tokenizer
    tokenizer_name = tokenizer_cfg['name']
    tokenizer_kwargs = tokenizer_cfg.get('kwargs', {})
    tokenizer = build_tokenizer(tokenizer_name, tokenizer_kwargs)

    evaluators, logger_keys, eval_gauntlet_callback = build_evaluators(
        eval_loader_config,
        icl_tasks,
        eval_gauntlet_config,
        tokenizer=tokenizer,
        device_eval_batch_size=device_eval_batch_size,
        icl_seq_len=max_seq_len,
        icl_subset_num_batches=icl_subset_num_batches,
    )

    # Callbacks
    callbacks: list[Callback] = [
        build_callback(name=str(name), kwargs=callback_cfg)
        for name, callback_cfg in callback_configs.items()
    ] if callback_configs else []

    if eval_gauntlet_callback is not None:
        callbacks.append(eval_gauntlet_callback)

    if metadata is not None:
        # Find the MosaicMLLogger
        mosaicml_logger = find_mosaicml_logger(loggers)

        if mosaicml_logger is not None:
            mosaicml_logger.log_metrics(metadata)
            mosaicml_logger._flush_metadata(force_flush=True)

    fsdp_config = parallelism_config.get(
        'fsdp',
        None,
    ) if parallelism_config else None
    if fsdp_config and model.get('load_in_8bit', False):
        raise ValueError(
            'The FSDP config block is not supported when loading ' +
            'Hugging Face models in 8bit.',
        )

    init_context = process_init_device(model, fsdp_config)

    name = model.pop('name')
    composer_model = build_composer_model(
        name=name,
        tokenizer=tokenizer,
        init_context=init_context,
        cfg=model,
    )

    # Now add the eval metrics
    if eval_loader_config is not None:
        train_metrics = composer_model.get_metrics(is_train=True)
        evaluators = add_metrics_to_eval_loaders(
            evaluators,
            list(train_metrics.keys()),
        )

    if eval_gauntlet_df is None and eval_gauntlet_callback is not None:
        eval_gauntlet_df = pd.DataFrame(
            columns=['model_name'] + list(eval_gauntlet_callback.averages) +
            [t['name'] for t in eval_gauntlet_callback.categories],
        )

    if name == 'mpt_causal_lm' and load_path is None:
        raise ValueError(
            'MPT causal LMs require a load_path to the checkpoint for model evaluation.'
            +
            ' Please check your yaml and the model_cfg to ensure that load_path is set.',
        )

    assert composer_model is not None

    log.info(f'Building trainer for {model_name}...')
    trainer = Trainer(
        run_name=run_name,
        seed=seed,
        model=composer_model,
        callbacks=callbacks,
        loggers=loggers,
        precision=precision,
        parallelism_config=parallelism_config,
        load_path=load_path,
        load_weights_only=True,
        progress_bar=False,
        log_to_console=True,
        dist_timeout=dist_timeout,
        python_log_level=python_log_level,
    )

    if should_log_config:
        log.info('Evaluation config:')
        log_config(trainer.logger, logged_config)

    log.info(f'Starting eval for {model_name}...')
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    a = time.time()
    trainer.eval(
        eval_dataloader=evaluators,
        subset_num_batches=eval_subset_num_batches,
    )
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    b = time.time()

    log.info(f'Ran {model_name} eval in: {b-a} seconds')
    return (trainer, logger_keys, eval_gauntlet_callback, eval_gauntlet_df)


def allow_toplevel_keys(cfg: dict[str, Any]) -> dict[str, Any]:
    """Transform the config to allow top-level keys for model configuration.

    This function allows users to use the 'train.py' syntax in 'eval.py'.
    It converts a config with top-level 'model', 'tokenizer', and (optionally) 'load_path' keys
    into the nested 'models' list format required by 'eval.py'.

    Input config format (train.py style):
    ```yaml
    model:
      <model_kwargs>
    load_path: /path/to/checkpoint
    tokenizer:
      <tokenizer_kwargs>
    ```

    Output config format (eval.py style):
    ```yaml
    models:
      - model:
          <model_kwargs>
        tokenizer:
          <tokenizer_kwargs>
        load_path: /path/to/checkpoint
    ```
    """
    if 'model' in cfg:
        if 'models' in cfg:
            raise ValueError(
                'Please specify either model or models in the config, not both',
            )
        default_name = cfg.get('model').get('name')  # type: ignore
        model_cfg = {
            'model': cfg.pop('model'),
            'tokenizer': cfg.pop('tokenizer', None),
            'model_name': cfg.pop('model_name', default_name),
        }
        if 'tokenizer' not in model_cfg or model_cfg['tokenizer'] is None:
            raise ValueError(
                'When specifying model, "tokenizer" must be provided in the config',
            )
        if 'load_path' in cfg:
            model_cfg['load_path'] = cfg.pop('load_path')
        cfg['models'] = [model_cfg]

    return cfg


def evaluate(cfg: DictConfig) -> tuple[list[Trainer], pd.DataFrame]:
    # Run user provided code if specified
    for code_path in cfg.get('code_paths', []):
        import_file(code_path)

    logged_cfg, eval_config = make_dataclass_and_log_config(
        cfg,
        EvalConfig,
        EVAL_CONFIG_KEYS,
        transforms=[allow_toplevel_keys],
        icl_tasks_required=False,
    )

    model_configs = eval_config.models
    eval_gauntlet_config = eval_config.eval_gauntlet or eval_config.eval_gauntlet_str

    # Mandatory Evaluation Parameters
    icl_tasks = eval_config.icl_tasks or eval_config.icl_tasks_str
    if icl_tasks is None:
        icl_tasks = []

    # Optional Evaluation Parameters with default values
    eval_loader_config = eval_config.eval_loader or eval_config.eval_loaders
    default_run_name: str = os.environ.get('RUN_NAME', 'llm')
    run_name = eval_config.run_name if eval_config.run_name else default_run_name

    reproducibility.seed_all(eval_config.seed)
    dist.initialize_dist(get_device(None), timeout=eval_config.dist_timeout)

    if eval_config.python_log_level is not None:
        logging.basicConfig(
            # Example of format string
            # 2022-06-29 11:22:26,152: rank0[822018][MainThread]: INFO: Message here
            format=
            f'%(asctime)s: rank{dist.get_global_rank()}[%(process)d][%(threadName)s]: %(levelname)s: %(name)s: %(message)s',
            force=True,
        )
        logging.getLogger('llmfoundry').setLevel(
            eval_config.python_log_level.upper(),
        )

    # Default argument values for evaluate_model
    eval_gauntlet_df = None
    models_df = None
    composite_scores = None
    trainers = []

    # Build loggers
    loggers: list[LoggerDestination] = [
        build_logger(name, logger_cfg)
        for name, logger_cfg in (eval_config.loggers or {}).items()
    ]

    mosaicml_logger = find_mosaicml_logger(loggers)
    if mosaicml_logger is None:
        mosaicml_logger = maybe_create_mosaicml_logger()
        # mosaicml_logger will be None if run isn't on MosaicML platform
        if mosaicml_logger is not None:
            loggers.append(mosaicml_logger)

    # mosaicml_logger will be None if the run isn't from the MosaicML platform
    if mosaicml_logger is not None:
        log_eval_analytics(
            mosaicml_logger,
            model_configs,
            icl_tasks,
            eval_gauntlet_config,
        )

    for model_cfg in model_configs:

        attn_config = model_cfg['model'].get('attn_config', None)
        if attn_config is not None:
            seq_parallel_world_size = attn_config.get(
                'seq_parallel_world_size',
                None,
            )
            if seq_parallel_world_size is not None and seq_parallel_world_size != 1:
                raise ValueError(
                    'Offline eval does not support sequence parallelism.',
                )

        (trainer, logger_keys, eval_gauntlet_callback,
         eval_gauntlet_df) = evaluate_model(
             dist_timeout=eval_config.dist_timeout,
             run_name=run_name,
             seed=eval_config.seed,
             icl_tasks=icl_tasks,
             max_seq_len=eval_config.max_seq_len,
             device_eval_batch_size=eval_config.device_eval_batch_size,
             eval_gauntlet_config=eval_gauntlet_config,
             eval_loader_config=eval_loader_config,
             loggers=loggers,
             python_log_level=eval_config.python_log_level,
             parallelism_config={'fsdp': eval_config.fsdp_config},
             precision=eval_config.precision,
             eval_gauntlet_df=eval_gauntlet_df,
             callback_configs=eval_config.callbacks,
             eval_subset_num_batches=eval_config.eval_subset_num_batches,
             icl_subset_num_batches=eval_config.icl_subset_num_batches,
             metadata=eval_config.metadata,
             logged_config=logged_cfg,
             should_log_config=eval_config.log_config,
             **model_cfg,
         )
        trainers.append(trainer)

        if eval_gauntlet_callback is not None:
            composite_scores = eval_gauntlet_callback.eval_after_all(
                trainer.state,
                trainer.logger,
            )

        benchmark_to_taxonomy = {}
        if eval_gauntlet_callback is not None:
            for t in eval_gauntlet_callback.categories:
                for b in t['benchmarks']:
                    benchmark_to_taxonomy[b['name']] = t['name']

        assert 'model_name' in model_cfg, 'model_name must be specified in model config'
        model_results = calculate_markdown_results(
            logger_keys,
            trainer,
            benchmark_to_taxonomy,
            model_cfg['model_name'],
        )

        if models_df is None:
            models_df = model_results
        else:
            models_df = pd.concat([models_df, model_results], ignore_index=True)

        if eval_gauntlet_df is not None and eval_gauntlet_callback is not None:
            assert composite_scores is not None
            row = {'model_name': model_cfg['model_name']}
            row.update({
                k.split('/')[-1]: v for k, v in composite_scores.items()
            })
            eval_gauntlet_df = pd.concat([
                eval_gauntlet_df,
                pd.DataFrame([row]),
            ],
                                         ignore_index=True)

            print(f'Printing gauntlet results for all models')

            print(
                eval_gauntlet_df.sort_values(
                    list(eval_gauntlet_callback.averages.keys())[0],
                    ascending=False,
                ).to_markdown(index=False),
            )
        print(f'Printing complete results for all models')
        assert models_df is not None
        print(models_df.to_markdown(index=False))

        trainer.close()

    return trainers, eval_gauntlet_df


def calculate_markdown_results(
    logger_keys: list[str],
    trainer: Trainer,
    benchmark_to_taxonomy: dict[str, str],
    model_name: str,
):
    results = {}

    for key in logger_keys:
        # dl_name is either 2-tuple (benchmark_name, num_fewshot)
        # or 3-tuple (benchmark_name, num_fewshot, subcategory)
        dl_name, metric_name = key.split('/')[1:-1], key.split('/')[-1]
        if 'Accuracy' not in metric_name:
            continue

        metric = trainer.state.eval_metrics.get('/'.join(dl_name),
                                                {}).get(metric_name, None)

        if metric is None:
            continue
        if dl_name[1] not in results:
            results[dl_name[1]] = {}

        if dl_name[0] not in results[dl_name[1]]:
            results[dl_name[1]][dl_name[0]] = {}

        if metric_name not in results[dl_name[1]][dl_name[0]]:
            results[dl_name[1]][dl_name[0]][metric_name] = []

        results[dl_name[1]][dl_name[0]][metric_name].append({
            'val': metric.compute(),
            'subcat': dl_name[-1] if len(dl_name) == 3 else 'no_subcat',
        })

    df = pd.DataFrame(
        columns=[
            'Category',
            'Benchmark',
            'Subtask',
            'Accuracy',
            'Number few shot',
            'Model',
        ],
    )

    for num_shot in results:
        for benchmark in results[num_shot]:
            for metric in results[num_shot][benchmark]:
                subscores = results[num_shot][benchmark][metric]
                if len(subscores) == 1:
                    row = {
                        'Category': benchmark_to_taxonomy.get(benchmark, ''),
                        'Benchmark': benchmark,
                        'Subtask': None,
                        'Accuracy': subscores[0]['val'],
                        'Number few shot': num_shot,
                        'Model': model_name,
                    }
                    df = pd.concat([df, pd.DataFrame([row])], ignore_index=True)
                else:
                    row = {
                        'Category':
                            benchmark_to_taxonomy.get(benchmark, ''),
                        'Benchmark':
                            benchmark,
                        'Subtask':
                            'Average',
                        'Accuracy':
                            sum(s['val'] for s in subscores) / len(subscores),
                        'Number few shot':
                            num_shot,
                        'Model':
                            model_name,
                    }
                    df = pd.concat([df, pd.DataFrame([row])], ignore_index=True)
                    for sub in subscores:
                        row = {
                            'Category':
                                benchmark_to_taxonomy.get(benchmark, ''),
                            'Benchmark':
                                None,
                            'Subtask':
                                sub['subcat'],
                            'Accuracy':
                                sub['val'],
                            'Number few shot':
                                num_shot,
                            'Model':
                                model_name,
                        }
                        df = pd.concat([df, pd.DataFrame([row])],
                                       ignore_index=True)
    return df


def eval_from_yaml(
    yaml_path: str,
    args_list: Optional[list[str]],
) -> tuple[list[Trainer], pd.DataFrame]:
    """Run the evaluation with optional overrides from CLI."""
    # Load yaml and CLI arguments.
    om.clear_resolver('oc.env')
    with open(yaml_path) as f:
        yaml_cfg = om.load(f)
    if args_list:
        cli_cfg = om.from_cli(args_list)
        yaml_cfg = om.merge(yaml_cfg, cli_cfg)
    assert isinstance(yaml_cfg, DictConfig)
    return evaluate(yaml_cfg)



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import contextlib
import copy
import logging
import math
import os
import warnings
from dataclasses import dataclass, fields
from pathlib import Path
from typing import (
    Any,
    Callable,
    Literal,
    Mapping,
    Optional,
    TypeVar,
    Union,
)

import mlflow
from composer.loggers import Logger
from composer.utils import dist, parse_uri
from mlflow.data import (
    delta_dataset_source,
    http_dataset_source,
    huggingface_dataset_source,
    uc_volume_dataset_source,
)
from omegaconf import MISSING, DictConfig, ListConfig, MissingMandatoryValue
from omegaconf import OmegaConf as om
from transformers import PretrainedConfig

from llmfoundry.layers_registry import ffns_with_megablocks
from llmfoundry.models.utils import init_empty_weights
from llmfoundry.registry import config_transforms

log = logging.getLogger(__name__)

__all__ = [
    'pop_config',
    'calculate_batch_size_info',
    'update_batch_size_info',
    'process_init_device',
    'log_config',
    'log_dataset_uri',
]


@dataclass
class EvalConfig:
    # Eval Config required parameters:
    models: list[dict[str, Any]] = MISSING
    max_seq_len: int = MISSING
    device_eval_batch_size: Union[int, float] = MISSING

    # Eval Config optional parameters:
    code_paths: Optional[list[str]] = None

    # Eval hyperparameters
    eval_gauntlet: Optional[dict[str, Any]] = None
    eval_gauntlet_str: Optional[str] = None
    eval_loader: Optional[dict[str, Any]] = None
    eval_loaders: Optional[list[dict[str, Any]]] = None
    eval_subset_num_batches: int = -1
    icl_subset_num_batches: Optional[int] = None
    # One of icl_tasks or icl_tasks_str must be specified
    icl_tasks: Optional[list[dict[str, Any]]] = None
    icl_tasks_str: Optional[str] = None

    # Logging parameters
    python_log_level: Optional[str] = 'debug'
    loggers: Optional[dict[str, Any]] = None
    console_log_interval: Union[int, str] = '1ba'
    log_config: bool = True

    # Model/run parameters
    seed: int = 17
    precision: str = 'amp_bf16'
    run_name: Optional[str] = None
    metadata: Optional[dict[str, str]] = None

    # Distributed parameters
    dist_timeout: Union[float, int] = 600.0
    fsdp_config: Optional[dict[str, Any]] = None

    # Callback parameters
    callbacks: Optional[dict[str, Any]] = None

    # Variables to ignore
    variables: Optional[dict[str, Any]] = None


EVAL_CONFIG_KEYS = {field.name for field in fields(EvalConfig)}


@dataclass
class TrainConfig:
    """Dataclass for training configuration."""

    # Mandatory model training parameters
    model: dict[str, Any] = MISSING
    tokenizer: dict[str, Any] = MISSING
    optimizer: dict[str, Any] = MISSING
    scheduler: dict[str, Any] = MISSING
    train_loader: dict[str, Any] = MISSING
    device_train_batch_size: Union[int, float] = MISSING
    device_eval_batch_size: Union[int, float] = MISSING
    max_duration: Union[int, str] = MISSING
    max_seq_len: int = MISSING

    # Seed
    seed: int = 17

    # Precision
    precision: str = 'amp_bf16'

    # Code paths to import
    code_paths: Optional[list[str]] = None

    # Cuda allocation configuration
    max_split_size_mb: Optional[int] = None
    expandable_segments: bool = True
    cuda_load_lazy: bool = False

    # Distributed training parameters
    dist_timeout: Union[int, float] = 600.0
    fsdp_config: Optional[dict[str, Any]] = None
    tp_config: Optional[dict[str, Any]] = None
    accumulate_train_batch_on_tokens: bool = True

    # Evaluation parameters
    eval_interval: Union[int, str] = 1
    eval_loader: Optional[dict[str, Any]] = None
    eval_loaders: Optional[list[dict[str, Any]]
                          ] = None  # should not be set by the user
    icl_tasks: Optional[list[dict[str, Any]]] = None
    icl_tasks_str: Optional[str] = None  # should not be set by the user
    eval_gauntlet: Optional[dict[str, Any]] = None
    eval_gauntlet_str: Optional[str] = None  # should not be set by the user
    icl_subset_num_batches: Optional[int] = None
    icl_seq_len: Optional[int] = None

    # Logging
    loggers: Optional[dict[str, Any]] = None
    progress_bar: bool = False
    log_to_console: bool = True
    python_log_level: Optional[str] = 'debug'
    console_log_interval: Union[int, str] = '1ba'
    log_config: bool = True

    # Callbacks
    callbacks: Optional[dict[str, Any]] = None
    algorithms: Optional[dict[str, Any]] = None

    # Checkpoints
    save_folder: Optional[str] = None
    save_latest_filename: Optional[str] = None
    save_overwrite: bool = False
    save_weights_only: bool = False
    save_filename: Optional[str] = None
    save_interval: Union[str, int] = '1000ba'
    save_num_checkpoints_to_keep: int = -1
    load_path: Optional[str] = None
    load_weights_only: bool = False
    load_strict_model_weights: bool = True
    load_ignore_keys: Optional[list[str]] = None
    save_ignore_keys: Optional[list[str]] = None
    only_hf_checkpoint: bool = False
    only_composer_checkpoint: bool = False

    # Dataloader
    train_subset_num_batches: int = -1
    device_train_microbatch_size: Union[str, int, float] = 'auto'
    global_train_batch_size: Optional[int] = None
    spin_dataloaders: bool = True

    # Eval dataloader
    eval_subset_num_batches: int = -1
    eval_first: bool = False
    compile_config: Optional[dict[str, Any]] = None

    # Metadata
    metadata: Optional[dict[str, Any]] = None
    flatten_metadata: bool = True
    run_name: Optional[str] = None

    # Resumption
    autoresume: bool = False

    # Profiling
    profiler: Optional[dict[str, Any]] = None

    # Variables to ignore
    variables: Optional[dict[str, Any]] = None

    # Fields created by `update_batch_size_info`
    n_gpus: int = MISSING
    device_train_grad_accum: Union[str, int] = MISSING


TRAIN_CONFIG_KEYS = {field.name for field in fields(TrainConfig)}


def forbid_config_key(cfg_dict: dict[str, Any], key: str):
    if key in cfg_dict:
        raise ValueError(
            f'Config key `{key}` should not be set. Please remove it from the config.',
        )


def to_dict_container(cfg: Union[DictConfig, dict[str, Any]]) -> dict[str, Any]:
    maybe_dict = to_container(cfg)
    if isinstance(maybe_dict, dict):
        return maybe_dict
    else:
        raise ValueError(f'Expected a dict-like type, got {type(maybe_dict)}')


def to_list_container(
    cfg: Union[ListConfig, list[dict[str, Any]]],
) -> list[dict[str, Any]]:
    maybe_list = to_container(cfg)
    if isinstance(maybe_list, list):
        return maybe_list
    else:
        raise ValueError(f'Expected a list-like type, got {type(maybe_list)}')


def to_container(
    cfg: Optional[Union[DictConfig, ListConfig, dict[str, Any],
                        list[dict[str, Any]]]],
) -> Union[dict[str, Any], list[dict[str, Any]]]:
    """Converts a DictConfig or ListConfig to a dict or list.

    `omegaconf.to_container` does not handle nested DictConfig or ListConfig
    objects, so this function is used to convert them to dicts or lists.
    """
    if isinstance(cfg, DictConfig):
        ret = om.to_container(cfg, resolve=True)
        assert isinstance(ret, dict)
        return ret  # type: ignore (return type is correct and converting all keys to str would be unnecessarily costly)
    elif isinstance(cfg, ListConfig):
        ret = om.to_container(cfg, resolve=True)
        assert isinstance(ret, list)
        return ret  # type: ignore (see above)
    else:
        return cfg  # type: ignore (dicts and lists are already in the correct format)


T = TypeVar('T')


def apply_transforms_to_config(
    cfg: dict[str, Any],
    transforms: Optional[Union[list[Callable[[dict[str, Any]], dict[str, Any]]],
                               list[str], str]],
) -> dict[str, Any]:
    """Applies a list of transforms to a config.

    Args:
        cfg (Dict[str, Any]): The config to transform.
        transforms (Optional[Union[List[Callable[[Dict[str, Any]], Dict[str, Any]]], List[str], str]]): A list of
            transform functions or strings representing transform functions to apply to the config. If a single string
            with the value ``all`` is provided, all registered transforms will be applied.

    Returns:
        Dict[str, Any]: The transformed config.
    """
    if transforms is None or (
        isinstance(transforms, list) and len(transforms) == 0
    ):
        return cfg

    transform_functions = []
    if isinstance(transforms, list):
        for transform in transforms:
            if isinstance(transform, str):
                transform_functions.append(config_transforms.get(transform))
            elif callable(transform):
                transform_functions.append(transform)
            else:
                raise ValueError(
                    f'Invalid transform: {transform}. Must be a string or callable.',
                )
    elif isinstance(transforms, str) and transforms == 'all':
        transform_functions = [
            config_transforms.get(transform)
            for transform in config_transforms.get_all()
        ]
    else:
        raise ValueError(
            f'Invalid transforms: {transforms}. Must be a list of strings or callables, or ``all``.',
        )

    for transform in transform_functions:
        cfg = transform(cfg)
    return cfg


def make_dataclass_and_log_config(
    cfg: DictConfig,
    dataclass_constructor: Callable[..., T],
    dataclass_fields: set[str],
    transforms: Optional[Union[list[Callable[[dict[str, Any]], dict[str, Any]]],
                               list[str], str]] = None,
    icl_tasks_required: bool = False,
) -> tuple[dict[str, Any], T]:
    """Converts a DictConfig to a dataclass and creates a logged config."""
    unstructured_config = om.to_container(cfg, resolve=True)
    assert isinstance(unstructured_config, dict)
    assert all(isinstance(k, str) for k in unstructured_config.keys())
    unstructured_config = {str(k): v for k, v in unstructured_config.items()}

    # Flatten union types before creating structured config:
    if 'eval_gauntlet' in unstructured_config:
        forbid_config_key(unstructured_config, 'eval_gauntlet_str')
        if isinstance(unstructured_config['eval_gauntlet'], str):
            unstructured_config['eval_gauntlet_str'] = unstructured_config.pop(
                'eval_gauntlet',
            )
    if (loader := unstructured_config.get('eval_loader', None)) is not None:
        forbid_config_key(unstructured_config, 'eval_loaders')
        if isinstance(loader, list):
            unstructured_config['eval_loaders'] = unstructured_config.pop(
                'eval_loader',
            )
    if 'icl_tasks' in unstructured_config:
        forbid_config_key(unstructured_config, 'icl_tasks_str')
        if isinstance(unstructured_config['icl_tasks'], str):
            unstructured_config['icl_tasks_str'] = unstructured_config.pop(
                'icl_tasks',
            )
    else:
        if icl_tasks_required:
            raise MissingMandatoryValue(
                'icl_tasks must be specified in the config',
            )

    # Apply transforms to the unstructured config before constructing dataclass
    unstructured_config = apply_transforms_to_config(
        unstructured_config,
        transforms,
    )

    # Create copy of config for logging
    logged_cfg: dict[str, Any] = copy.deepcopy(unstructured_config)

    arg_config_keys = set(unstructured_config.keys())
    extraneous_keys = set.difference(arg_config_keys, dataclass_fields)

    if 'variables' not in unstructured_config:
        unstructured_config['variables'] = {}

    if len(extraneous_keys) > 0:
        raise ValueError(
            f'Unused parameters {sorted(extraneous_keys)} found in cfg. Please check your yaml to ensure these parameters are necessary. Please place any variables under the `variables` key.',
        )

    dataclass_dict_config: DictConfig = om.structured(
        dataclass_constructor(**unstructured_config),
    )

    # Error on missing mandatory values:
    for key in dataclass_fields:
        _ = dataclass_dict_config[key]

    # Convert DictConfig to dict for dataclass constructor so that child
    # configs are not DictConfigs
    dataclass_config: T = dataclass_constructor(
        **to_dict_container(dataclass_dict_config),
    )

    return logged_cfg, dataclass_config


def pop_config(
    cfg: Union[dict[str, Any], DictConfig],
    key: str,
    must_exist: bool = True,
    default_value: Any = None,
    convert: bool = False,
) -> Any:
    """Pop a value from the main config file and return it.

    If the key does not exist, return the default_value or raise a RuntimeError
    depending on the must_exist flag. If the convert flag is set to True, then
    we will convert the value to a python object using OmegaConf.to_container.
    """
    value = cfg.pop(key, None)
    if value is not None and convert:
        if not isinstance(value,
                          DictConfig) and not isinstance(value, ListConfig):
            raise ValueError(
                f'The key {key} has a value of type {type(value)} that cannot be \
                            converted to a dict or list. Please check your yaml.',
            )
        return om.to_container(value)
    elif value is not None:
        return value
    elif must_exist:
        raise NameError(
            f'The {key} parameter is missing and must exist for execution. Please check your yaml.',
        )
    else:
        return default_value


def get_hf_config_value(config: Union[dict, PretrainedConfig], key: str) -> Any:
    """Get a value from a Hugging Face config.

    Args:
        config (Union[dict, PretrainedConfig]): The Hugging Face config object.
        key (str): The key to get from the config.

    Returns:
        Any: The value from the config. None if the key does not exist.
    """
    if isinstance(config, dict):
        return config.get(key)
    return getattr(config, key, None)


def calculate_batch_size_info(
    global_batch_size: int,
    device_microbatch_size: Union[int, float, Literal['auto']],
    data_replication_degree: int = 1,
) -> tuple[Union[int, float], Union[int, float, Literal['auto']], Union[
    int, Literal['auto']]]:

    world_size = dist.get_world_size()
    if world_size % data_replication_degree != 0:
        raise ValueError(
            f'World size {world_size} is not divisible by data replication degree {data_replication_degree}.',
        )
    if global_batch_size % (world_size // data_replication_degree) != 0:
        raise ValueError(
            f'Global batchsize {global_batch_size} is not divisible by {(world_size // data_replication_degree)=} '
            +
            'as a result, the batch size would be truncated, please adjust `global_batch_size` '
            + f'to be divisible by world size, {world_size}.',
        )
    device_batch_size = global_batch_size / world_size
    if device_batch_size == round(device_batch_size):
        device_batch_size = round(device_batch_size)
    if device_microbatch_size == 'auto':
        device_grad_accum = 'auto'
    elif isinstance(device_microbatch_size, (int, float)):
        if device_microbatch_size > device_batch_size:
            log.warn(
                f'device_microbatch_size > device_batch_size, ' +
                f'will be reduced from {device_microbatch_size} -> {device_batch_size}.',
            )
            device_microbatch_size = device_batch_size
        device_grad_accum = math.ceil(
            device_batch_size / device_microbatch_size,
        )
    else:
        raise ValueError(f'Not sure how to parse {device_microbatch_size=}')

    return device_batch_size, device_microbatch_size, device_grad_accum


def update_config_with_batch_size_info(
    cfg: dict[str, Any],
    device_train_batch_size: Union[int, float],
    device_train_microbatch_size: Union[int, float, Literal['auto']],
    device_train_grad_accum: Union[int, Literal['auto']],
) -> dict[str, Any]:
    """Update the config with batch size information.

    Args:
        cfg (Dict[str, Any]): The config to update.
        device_train_batch_size (Union[int, float]): The batch size of the training dataset for each device.
        device_train_microbatch_size (Union[int, float, Literal['auto']]): The microbatch size of the training dataset for each device.
        device_train_grad_accum (Union[int, Literal['auto']]): The gradient accumulation settings for each device.

    Returns:
        Dict[str, Any]: The updated config.
    """
    cfg['n_gpus'] = dist.get_world_size()
    cfg['device_train_batch_size'] = device_train_batch_size
    cfg['device_train_microbatch_size'] = device_train_microbatch_size
    cfg['device_train_grad_accum'] = device_train_grad_accum
    # Safely set `device_eval_batch_size` if not provided by user
    if 'device_eval_batch_size' not in cfg:
        if cfg['device_train_microbatch_size'] == 'auto':
            cfg['device_eval_batch_size'
               ] = 1  # TODO debug auto eval microbatching
        else:
            cfg['device_eval_batch_size'] = cfg['device_train_microbatch_size']
    return cfg


def update_batch_size_info(cfg: dict[str, Any]) -> dict[str, Any]:
    data_replication_degree = 1
    device_train_batch_size, device_train_microbatch_size, device_train_grad_accum = calculate_batch_size_info(
        cfg['global_train_batch_size'],
        cfg['device_train_microbatch_size'],
        data_replication_degree=data_replication_degree,
    )
    cfg = update_config_with_batch_size_info(
        cfg,
        device_train_batch_size,
        device_train_microbatch_size,
        device_train_grad_accum,
    )
    return cfg


def process_init_device(
    model_cfg: dict[str, Any],
    fsdp_config: Optional[dict] = None,
    tp_config: Optional[dict] = None,
):
    # Restrict model init_device to 'meta' and 'cpu',
    # using 'cuda' vs. 'cuda:id' is tricky and can lead to common user errors
    # when multiple GPUs are available.
    # Also 'meta' is only valid when using FSDP
    init_context = contextlib.nullcontext()
    if 'init_device' in model_cfg:
        assert model_cfg['init_device'] in ['meta', 'cpu', 'mixed']
        if fsdp_config is None and model_cfg['init_device'] == 'meta':
            warnings.warn(
                "Using `cfg.model.init_device='meta'` is only valid when using FSDP! " +\
                "Reverting to `cfg.model.init_device='cpu'`.")
            model_cfg['init_device'] = 'cpu'
        if model_cfg['init_device'] == 'meta':
            init_context = init_empty_weights()
        if model_cfg['init_device'] == 'mixed':
            if fsdp_config is None:
                raise NotImplementedError(
                    'Using init_device `mixed` is only supported with FSDP. ' +
                    'Please add a FSDP config.',
                )
            # Always set `sync_module_states` to True for mixed initialization
            if not fsdp_config.get('sync_module_states', False):
                warnings.warn((
                    'Setting `sync_module_states = True` for FSDP. This is required '
                    'when using mixed initialization.'
                ))
                fsdp_config['sync_module_states'] = True

            # Set defaults for mixed initialization
            fsdp_config.setdefault('load_monolith_rank0_only', True)

    if tp_config is not None:
        # Check tp_config has required fields
        if 'strategy' not in tp_config or 'tensor_parallel_degree' not in tp_config:
            raise ValueError(
                "`tp_config` requires 'strategy' and 'tensor_parallel_degree' values. ",
            )

        # Check we are not using tensor parallelism with MoEs
        if 'ffn_config' in model_cfg and model_cfg['ffn_config'].get(
            'ffn_type',
            None,
        ) in ffns_with_megablocks:
            raise ValueError(
                'Tensor Parallelism is not currently supported for MoE models.',
            )

    # Check we are not using tensor parallelism with MoEs
    if tp_config is not None and 'ffn_config' in model_cfg and model_cfg[
        'ffn_config'].get('ffn_type', None) in ffns_with_megablocks:
        raise ValueError(
            'Tensor Parallelism is not currently supported for MoE models.',
        )

    # Set ffn_config.device_mesh using fsdp_config
    if fsdp_config is not None and 'ffn_config' in model_cfg and model_cfg[
        'ffn_config'].get('ffn_type', None) in ffns_with_megablocks:
        shard_degree = fsdp_config.get('data_parallel_shard_degree', None)
        replicate_degree = fsdp_config.get(
            'data_parallel_replicate_degree',
            None,
        )

        if shard_degree is None and replicate_degree is None:
            # Default to sharding over all gpus.
            shard_degree = dist.get_world_size()
            device_mesh_cfg = [shard_degree]
        else:
            if shard_degree is None:
                # Shard degree is not specified, so calculate it from replicate degree
                assert isinstance(replicate_degree, int)
                shard_degree = dist.get_world_size() // replicate_degree
            elif replicate_degree is None:
                # Replicate degree is not specified, so calculate it from shard degree
                assert isinstance(shard_degree, int)
                replicate_degree = dist.get_world_size() // shard_degree

            if replicate_degree == 1:
                device_mesh_cfg = [shard_degree]
            else:
                device_mesh_cfg = [replicate_degree, shard_degree]

        model_cfg['ffn_config']['device_mesh'] = device_mesh_cfg

    # No mixed precision needed for weights when they're already 16 bits
    master_dtype = model_cfg.get('master_weights_dtype')
    small_dtypes = (
        'bf16',
        'fp16',
        'float16',
        'bfloat16',
        'amp_fp16',
        'amp_bf16',
    )
    if fsdp_config and master_dtype in small_dtypes:
        reduce_dtype = None
        buffer_dtype = None
        mixed_precision = fsdp_config.get('mixed_precision')
        if isinstance(mixed_precision, Mapping):
            reduce_dtype = mixed_precision.get('reduce_dtype')
            buffer_dtype = mixed_precision.get('buffer_dtype')
        fsdp_config['mixed_precision'] = {
            'param_dtype': None,
            'reduce_dtype': reduce_dtype,
            'buffer_dtype': buffer_dtype,
            'keep_low_precision_grads': True,
        }

    return init_context


def log_config(logger: Logger, cfg: dict[str, Any]) -> None:
    """Logs the current config and updates the wandb and mlflow configs.

    This function can be called multiple times to update the wandb and MLflow
    config with different variables.
    """
    print(om.to_yaml(cfg))
    logger.log_hyperparameters(cfg)


def _parse_source_dataset(cfg: dict[str, Any]) -> list[tuple[str, str, str]]:
    """Parse a run config for dataset information.

    Given a config dictionary, parse through it to determine what the datasource
    should be categorized as. Possible data sources are Delta Tables, UC Volumes,
    HuggingFace paths, remote storage, or local storage.

    Args:
        cfg (DictConfig): A config dictionary of a run

    Returns:
        List[Tuple[str, str, str]]: A list of tuples formatted as (data type, path, split)
    """
    data_paths = []

    # Handle train loader if it exists
    train_dataset: dict = cfg.get('train_loader', {}).get('dataset', {})
    train_split = train_dataset.get('split', None)
    train_source_path = cfg.get('source_dataset_train', None)
    _process_data_source(
        train_source_path,
        train_dataset,
        train_split,
        'train',
        data_paths,
    )

    # Handle eval_loader which might be a list or a single dictionary
    eval_data_loaders = cfg.get('eval_loader', {})
    if not isinstance(eval_data_loaders, list):
        eval_data_loaders = [
            eval_data_loaders,
        ]  # Normalize to list if it's a single dictionary

    for eval_data_loader in eval_data_loaders:
        assert isinstance(eval_data_loader, dict)  # pyright type check
        eval_dataset: dict = eval_data_loader.get('dataset', {})
        eval_split = eval_dataset.get('split', None)
        eval_source_path = cfg.get('source_dataset_eval', None)
        _process_data_source(
            eval_source_path,
            eval_dataset,
            eval_split,
            'eval',
            data_paths,
        )

    return data_paths


def _process_data_source(
    source_dataset_path: Optional[str],
    dataset: dict[str, str],
    cfg_split: Optional[str],
    true_split: str,
    data_paths: list[tuple[str, str, str]],
):
    """Add a data source by mutating data_paths.

    Given various dataset attributes, attempt to determine what type of dataset is being added, and parse
    the dataset accordingly.

    Args:
        source_dataset_path (Optional[str]): The source dataset in cfg metadata
        dataset (Dict[str, str]): The dataset from cfg
        cfg_split (str): The split listed for the dataset in cfg
        true_split (str): The split of the dataset to be added (i.e. train or eval)
        data_paths (List[Tuple[str, str, str]]): A list of tuples formatted as (data type, path, split)
    """
    if source_dataset_path:
        source_dataset_path = str(Path(source_dataset_path))
    # Check for Delta table
    if source_dataset_path and len(source_dataset_path.split('.')) == 3:
        data_paths.append(('delta_table', source_dataset_path, true_split))
    # Check for UC volume
    elif source_dataset_path and source_dataset_path.startswith('dbfs:'):
        data_paths.append(
            ('uc_volume', source_dataset_path[len('dbfs:'):], true_split),
        )
    # Check for HF path
    elif 'hf_name' in dataset and dataset['hf_name']:
        hf_path = dataset['hf_name']
        backend, _, uc_path = parse_uri(hf_path)
        unsupported_file = True
        if backend == 'dbfs':
            assert cfg_split
            from llmfoundry.data.finetuning.tasks import SUPPORTED_EXTENSIONS
            possible_files = [
                f'{cfg_split}{ext}' for ext in SUPPORTED_EXTENSIONS
            ]
            for file in possible_files:
                path = os.path.join(uc_path, file)
                # Ensure path starts with '/'
                if not path.startswith('/'):
                    path = '/' + path
                if _verify_uc_path(path):
                    data_paths.append(('uc_volume', path, true_split))
                    unsupported_file = False
                    break
            if unsupported_file:
                log.warning(
                    f'{hf_path} does not contain a supported file extension.',
                )
        elif backend:
            hf_path = os.path.join(hf_path, cfg_split) if cfg_split else hf_path
            data_paths.append((backend, hf_path, true_split))
        elif os.path.exists(hf_path):
            data_paths.append(('local', hf_path, true_split))
        else:
            data_paths.append(('hf', hf_path, true_split))
    # Check for remote path
    elif 'remote' in dataset and dataset['remote']:
        remote_path = dataset['remote']
        backend, _, _ = parse_uri(remote_path)
        if backend:
            remote_path = os.path.join(
                remote_path,
                f'{cfg_split}/',
            ) if cfg_split else remote_path
            data_paths.append((backend, remote_path, true_split))
        else:
            # No backend detected so assume local path
            data_paths.append(('local', remote_path, true_split))
    # Check for local path
    elif 'local' in dataset and dataset['local']:
        data_paths.append(('local', dataset['local'], true_split))
    else:
        log.warning('DataSource Not Found.')


def log_dataset_uri(cfg: dict[str, Any]) -> None:
    """Logs dataset tracking information to MLflow.

    Args:
        cfg (DictConfig): A config dictionary of a run
    """
    loggers = cfg.get('loggers', None) or {}
    if 'mlflow' not in loggers or not mlflow.active_run():
        return
    # Figure out which data source to use
    data_paths = _parse_source_dataset(cfg)

    dataset_source_mapping = {
        's3': http_dataset_source.HTTPDatasetSource,
        'oci': http_dataset_source.HTTPDatasetSource,
        'azure': http_dataset_source.HTTPDatasetSource,
        'gs': http_dataset_source.HTTPDatasetSource,
        'https': http_dataset_source.HTTPDatasetSource,
        'hf': huggingface_dataset_source.HuggingFaceDatasetSource,
        'delta_table': delta_dataset_source.DeltaDatasetSource,
        'uc_volume': uc_volume_dataset_source.UCVolumeDatasetSource,
        'local': http_dataset_source.HTTPDatasetSource,
    }

    # Map data source types to their respective MLFlow DataSource.
    for dataset_type, path, split in data_paths:
        if dataset_type in dataset_source_mapping:
            source_class = dataset_source_mapping[dataset_type]
            if dataset_type == 'delta_table':
                source = source_class(delta_table_name=path)
            elif dataset_type == 'hf' or dataset_type == 'uc_volume':
                source = source_class(path=path)
            else:
                source = source_class(url=path)
        else:
            log.info(
                f'{dataset_type} unknown, defaulting to http dataset source',
            )
            source = http_dataset_source.HTTPDatasetSource(url=path)

        mlflow.log_input(
            mlflow.data.meta_dataset.MetaDataset(source, name=split),
        )


def _verify_uc_path(path: str) -> bool:
    """Verify a UC path exists.

    Args:
        path (str): UnityCatalog path
    Returns:
        (bool): If path exists or not
    """
    from databricks.sdk.errors.platform import NotFound, PermissionDenied
    w = None
    try:
        from databricks.sdk import WorkspaceClient

        w = WorkspaceClient()
    except ImportError:
        log.warning(
            'Cannot verify the path of `UCVolumeDatasetSource` because of missing' + \
            '`databricks-sdk`. Please install `databricks-sdk` via ' + \
            '`pip install -U databricks-sdk`. This does not block creating ' + \
            '`UCVolumeDatasetSource`, but your `UCVolumeDatasetSource` might be invalid.',
        )
        return False
    except Exception as e:
        log.warning(
            f'Error occured when attempting to connect with Databricks WorkspaceClient. ' + \
            f'Error details: {str(e)}. This does not block creating `UCVolumeDatasetSource`, ' + \
            f'but your `UCVolumeDatasetSource` might be invalid.',
        )

    if w:
        try:
            w.files.get_metadata(path)
        except (NotFound, PermissionDenied):
            try:
                # Check if `self.path` points to a valid UC directory.
                w.files.get_directory_metadata(path)
                return True
            except (NotFound, PermissionDenied):
                # Neither file nor directory exists, we throw an exception.
                return False
        except Exception as e:
            log.warning(
                f'Error occured when verifying path of `UCVolumeDatasetSource`. ' + \
                f'Error details: {str(e)}. This does not block creating `UCVolumeDatasetSource`, ' + \
                f'but your `UCVolumeDatasetSource` might be invalid.',
            )
    return False


def set_config_overrides(
    config: PretrainedConfig,
    config_overrides: dict[str, Any],
):
    # set config overrides
    for k, v in config_overrides.items():
        if not hasattr(config, k):
            raise ValueError(
                f'config does not have attribute "{k}" to override ({k}: {v}).',
            )

        attr = getattr(config, k)
        # attempt to disallow typos in nested configs
        if isinstance(attr, Mapping):
            extra_keys = [_k for _k in v.keys() if _k not in attr.keys()]
            if extra_keys:
                raise ValueError(
                    f'Config dict override got unknown keys. ' +
                    f'Extra keys: {extra_keys}. ' +
                    f'Expected (a subset of) keys: {list(attr.keys())}.',
                )
            getattr(config, k).update(v)
        # necessary case to allow for rope_scaling to be overriden in llama config
        elif attr is None and isinstance(v, Mapping):
            setattr(config, k, {})
            getattr(config, k).update(v)
        elif isinstance(attr, PretrainedConfig):
            if not isinstance(v, Mapping):
                raise ValueError(
                    f'Expected a dictionary for config override {k}, but got {v}.',
                )

            for _k, _v in v.items():
                if not hasattr(attr, _k):
                    raise ValueError(
                        f'config does not have attribute "{_k}" to override ({k}: {_k}: {_v}).',
                    )
                setattr(attr, _k, _v)
        else:
            setattr(config, k, v)



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry.eval.datasets.in_context_learning_evaluation import (
    InContextLearningDataset,
    InContextLearningGenerationTaskWithAnswersDataset,
    InContextLearningLMTaskDataset,
    InContextLearningMultipleChoiceTaskDataset,
    InContextLearningSchemaTaskDataset,
    get_icl_task_dataloader,
)
from llmfoundry.eval.metrics.nlp import (
    InContextLearningGenerationExactMatchAccuracy,
    InContextLearningLMAccuracy,
    InContextLearningLMExpectedCalibrationError,
    InContextLearningMCExpectedCalibrationError,
    InContextLearningMetric,
    InContextLearningMultipleChoiceAccuracy,
)

__all__ = [
    'InContextLearningDataset',
    'InContextLearningLMTaskDataset',
    'InContextLearningMultipleChoiceTaskDataset',
    'InContextLearningSchemaTaskDataset',
    'InContextLearningGenerationTaskWithAnswersDataset',
    'get_icl_task_dataloader',
    'InContextLearningMetric',
    'InContextLearningLMAccuracy',
    'InContextLearningMultipleChoiceAccuracy',
    'InContextLearningGenerationExactMatchAccuracy',
    'InContextLearningLMExpectedCalibrationError',
    'InContextLearningMCExpectedCalibrationError',
]



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""A collection of common torchmetrics for NLP tasks."""

import copy
import functools
import logging
import re
import string
from typing import Any, Callable

import torch
from torch import Tensor
from torch.nn import functional as F
from torchmetrics import Metric

log = logging.getLogger(__name__)

__all__ = [
    'InContextLearningMetric',
    'InContextLearningLMAccuracy',
    'InContextLearningMultipleChoiceAccuracy',
    'InContextLearningGenerationExactMatchAccuracy',
    'InContextLearningLMExpectedCalibrationError',
    'InContextLearningMCExpectedCalibrationError',
]


class InContextLearningMetric(Metric):

    def __init__(self, *args, **kwargs):  # pyright: ignore
        super().__init__(*args, **kwargs)
        self.needs_batch = True

    def _wrap_update(self, update: Callable) -> Callable:
        """Overwrite default _wrap_update to return result of update().

        Torch metrics wraps update with following wrapped_func but explicitly
        does not return the value. In general, torchmetrics update() does not
        return a value, but we want to in order to pass it on to
        state.metric_outputs.
        """

        @functools.wraps(update)
        def wrapped_func(*args: Any, **kwargs: Any) -> None:
            self._computed = None
            self._update_count += 1
            with torch.set_grad_enabled(self._enable_grad):
                try:
                    update_result = update(*args, **kwargs)
                except RuntimeError as err:
                    if 'Expected all tensors to be on' in str(err):
                        raise RuntimeError(
                            'Encountered different devices in metric calculation (see stacktrace for details).' + \
                            ' This could be due to the metric class not being on the same device as input.' + \
                            f' Instead of `metric={self.__class__.__name__}(...)` try to do' + \
                            f' `metric={self.__class__.__name__}(...).to(device)` where' + \
                            ' device corresponds to the device of the input.',
                        ) from err
                    raise err

            if self.compute_on_cpu:
                self._move_list_states_to_cpu()
            return update_result

        return wrapped_func

    def update(
        self,
        batch: dict,
        outputs: torch.Tensor,
        labels: torch.Tensor,
    ):
        """Abstract interface for computing an in-context learning metrics.

        The `output_logits` argument is deprecated and will be removed in v0.21 while it's functionality will
        be moved to `outputs`.

        Args:
            batch (dict): Batch must consist minimally of `input_ids` as well as any other structure needed
                to compute the metric.
            outputs (torch.Tensor): The model outputs evaluated on the batch `input_ids`.
            labels (torch.Tensor): The correct outputs.

        Raises:
            NotImplementedError: Abstract method must be implemented by subclasses
        """
        raise NotImplementedError


class InContextLearningGenerationExactMatchAccuracy(InContextLearningMetric):
    r"""Computes exact match for in-context learning generation tasks.

    ICL generation tasks consist of some number of prompted generation tasks with correct answers
    followed by a test task where the model must correctly produce one of a number of valid answers.

    For example, the model may be provided the context below and evaluated on its ability to correctly predict the continuation.

    Context: `Question: Who was president of the United States in 2012?\nAnswer: Barack Obama\nQuestion: Is water wet?\nAnswer: `
    Answers: [`yes`]

    The model will be expected to correctly produce one of the answers, following some optional normalization.

    Adds metric state variables:
        correct (float): The number of instances where the prediction was a prefix for any of the answer aliases.
        total (float): The number of total instances that were predicted.

    Args:
        dist_sync_on_step (bool, optional): Synchronize metric state across processes at
            each forward() before returning the value at the step. Default: ``False``.
    """

    # Make torchmetrics call update only once
    full_state_update = False

    def __init__(self, dist_sync_on_step: bool = False):
        # state from multiple processes
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state(
            'correct',
            default=torch.tensor(0.),
            dist_reduce_fx='sum',
        )
        self.add_state('total', default=torch.tensor(0.), dist_reduce_fx='sum')
        self.metric_result_dict = {
            'cleaned_output': [],
            'original_label': [],
            'cleaned_label': [],
            'result': [],
        }

    def normalize_answer(self, answer: str):
        """Lower text and remove punctuation, articles and extra whitespace.

        Copied from https://github.com/mandarjoshi90/triviaqa/blob/master/evaluation/triviaqa_evaluation.py
        """

        def remove_articles(text: str) -> str:
            return re.sub(r'\b(a|an|the)\b', ' ', text)

        def white_space_fix(text: str) -> str:
            return ' '.join(text.split())

        def handle_punc(text: str) -> str:
            exclude = set(
                string.punctuation + ''.join([u'‘', u'’', u'´', u'`']),
            )
            return ''.join(ch if ch not in exclude else ' ' for ch in text)

        def lower(text: str) -> str:
            return text.lower()

        def replace_underscore(text: str) -> str:
            return text.replace('_', ' ')

        return white_space_fix(
            remove_articles(handle_punc(lower(replace_underscore(answer)))),
        ).strip()

    def update(
        self,
        batch: dict[str, Any],
        outputs: list[str],
        labels: list[list[str]],
    ):
        cot_delimiter = batch.get('cot_delimiter', '')
        do_normalization = batch.get('do_normalization', True)
        stopping_criteria = batch.get('stopping_criteria', None)
        metric_result_dict = copy.deepcopy(self.metric_result_dict)
        for sample_output, sample_labels in zip(outputs, labels):
            final_answer = sample_output

            if stopping_criteria is not None and len(stopping_criteria) > 0:
                final_answer = re.split(
                    '|'.join(stopping_criteria),
                    final_answer,
                )[0]

            if cot_delimiter is not None and len(cot_delimiter) > 0:
                final_answer = final_answer.split(cot_delimiter)[-1]

            if do_normalization:
                cleaned_final_answer = self.normalize_answer(final_answer)
                cleaned_sample_labels = {
                    self.normalize_answer(label) for label in sample_labels
                }
            else:
                # even if normalization is off, we should still strip leading/trailing whitespaces
                cleaned_final_answer = final_answer.strip()
                cleaned_sample_labels = {
                    sample_label.strip() for sample_label in sample_labels
                }
            metric_result_dict['original_label'].append(sample_labels)
            metric_result_dict['cleaned_output'].append(cleaned_final_answer)
            metric_result_dict['cleaned_label'].append(cleaned_sample_labels)

            if any(
                cleaned_final_answer.startswith(label)
                for label in cleaned_sample_labels
            ):
                self.correct += torch.tensor(1.0)
                metric_result_dict['result'].append(1)
            else:
                metric_result_dict['result'].append(0)

            self.total += torch.tensor(1.0)

        return metric_result_dict

    def compute(self):
        assert isinstance(self.correct, Tensor)
        assert isinstance(self.total, Tensor)
        return self.correct / self.total


class InContextLearningLMAccuracy(InContextLearningMetric):
    r"""Computes accuracy for In-context learning language modeling tasks.

    ICL LM tasks consist of some number of example language modeling tasks (referred to as the 'context'), followed by a test task where the model must correctly predict all the tokens
    following tokens in some passage (referred to as the 'continuation').

    For example, the model may be provided the context below and evaluated on its ability to correctly predict the continuation. Note: it doesn't matter
    whether the model correctly predicts the context tokens.

    Context: `The dog is->fuzzy\nthe water is->hot\nthe tree is->`
    Continuation: `green`

    Adds metric state variables:
        correct (float): The number of instances where the prediction masked the target.
        total (float): The number of total instances that were predicted.

    Args:
        dist_sync_on_step (bool, optional): Synchronize metric state across processes at
            each forward() before returning the value at the step. Default: ``False``.
    """

    # Make torchmetrics call update only once
    full_state_update = False

    def __init__(self, dist_sync_on_step: bool = False):
        # state from multiple processes
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state(
            'correct',
            default=torch.tensor(0.),
            dist_reduce_fx='sum',
        )
        self.add_state('total', default=torch.tensor(0.), dist_reduce_fx='sum')
        self.metric_result_dict = {
            'context': [],
            'label': [],
            'output': [],
            'result': [],
        }

    def update(self, batch: dict, outputs: torch.Tensor, labels: torch.Tensor):

        metric_result_dict = copy.deepcopy(self.metric_result_dict)
        for batch_idx, cont_idx in enumerate(batch['continuation_indices']):
            cont_tok_pred = outputs[batch_idx].index_select(
                dim=0,
                index=cont_idx - 1,
            ).argmax(dim=-1)
            cont_tok_targ = labels[batch_idx].index_select(
                dim=0,
                index=cont_idx - 1,
            )

            metric_result_dict['context'].append(
                batch['input_ids'][batch_idx][:cont_idx[0]],
            )
            metric_result_dict['label'].append(cont_tok_targ)
            metric_result_dict['output'].append(cont_tok_pred)

            correct = (cont_tok_pred == cont_tok_targ).all().int()
            self.correct += correct
            metric_result_dict['result'].append(int(correct))

            self.total += torch.tensor(1.0)

        return metric_result_dict

    def compute(self):
        assert isinstance(self.correct, Tensor)
        assert isinstance(self.total, Tensor)
        return self.correct / self.total


class InContextLearningMultipleChoiceAccuracy(InContextLearningMetric):
    r"""Computes accuracy for In-context learning multiple choice tasks.

    ICL MC tasks consists of a series of questions with some number of possible choices (only one of which can be correct).
    At inference time each possible choice is given to the model as a separate input and the one for which the model assigns
    the lowest perplexity to the choice is considered the model's choice. The model is correct if it "chooses" the right answer.

    Context: `The dog is->fuzzy\nthe water is->hot\nthe tree is->`
    Continuation: `green`

    Adds metric state variables:
        correct (float): The number of instances where the prediction masked the target.
        total (float): The number of total instances that were predicted.

    Args:
        dist_sync_on_step (bool, optional): Synchronize metric state across processes at
            each forward() before returning the value at the step. Default: ``False``.
    """

    # Make torchmetrics call update only once
    full_state_update = False

    def __init__(self, dist_sync_on_step: bool = False):
        # state from multiple processes
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state(
            'correct',
            default=torch.tensor(0.0),
            dist_reduce_fx='sum',
        )
        self.add_state('total', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.metric_result_dict = {
            'context': [],
            'correct_choice': [],
            'correct_choice_idx': [],
            'selected_choice': [],
            'selected_choice_idx': [],
            'all_choices': [],
            'result': [],
        }

    def update(self, batch: dict, outputs: torch.Tensor, labels: torch.Tensor):

        perplexities = []
        for batch_idx, cont_idx in enumerate(batch['continuation_indices']):
            # continuation indices refer to indices in the original input's token space
            cont_tok_logits = outputs[batch_idx].index_select(
                dim=0,
                index=cont_idx - 1,
            )
            # labels have been shifted left by one index, so the cont_idx needs to be shifted as well.
            cont_tok_targ = labels[batch_idx].index_select(
                dim=0,
                index=cont_idx - 1,
            )
            cross_entropy = F.cross_entropy(cont_tok_logits, cont_tok_targ)
            perplexity = torch.exp(cross_entropy)
            perplexities.append(perplexity)

        metric_result_dict = copy.deepcopy(self.metric_result_dict)
        for (start, end), gold_idx in zip(
            batch['choice_groupings'],
            batch['gold_indices'],
        ):
            subset = perplexities[start:end]
            idx_min = subset.index(min(subset))

            if idx_min == gold_idx:
                self.correct += torch.tensor(1.0)
                metric_result_dict['result'].append(1)
            else:
                metric_result_dict['result'].append(0)

            question = batch['input_ids'][
                start][:batch['continuation_indices'][start][0]]

            correct_choice = batch['input_ids'][start:end][gold_idx][
                batch['continuation_indices'][start:end][gold_idx][0]:
                batch['continuation_indices'][start:end][gold_idx][-1] + 1]
            selected_choice = batch['input_ids'][start:end][idx_min][
                batch['continuation_indices'][start:end][idx_min][0]:
                batch['continuation_indices'][start:end][idx_min][-1] + 1]
            metric_result_dict['context'].append(question)
            metric_result_dict['correct_choice'].append(correct_choice)
            metric_result_dict['correct_choice_idx'].append(gold_idx)
            metric_result_dict['selected_choice'].append(selected_choice)
            metric_result_dict['selected_choice_idx'].append(idx_min)
            all_choices = batch['input_ids'][start:end]
            # Unpads the choices. Necessary in case different choices have different token lengths.
            if 'attention_mask' in batch:
                all_choices_list = [
                    choice[batch['attention_mask'][i]]
                    for i, choice in enumerate(all_choices)
                ]
                metric_result_dict['all_choices'].append(all_choices_list)

            self.total += torch.tensor(1.0)

        # Don't return all_choices if we didn't fill it up (i.e. didn't use causal lms)
        if metric_result_dict['all_choices'] == []:
            metric_result_dict.pop('all_choices')

        return metric_result_dict

    def compute(self):
        assert isinstance(self.correct, Tensor)
        assert isinstance(self.total, Tensor)
        return self.correct.float() / self.total


class InContextLearningExpectedCalibrationError(InContextLearningMetric):
    """Generic class for Expected Calibration Error (ECE).

    Citation: https://arxiv.org/pdf/1706.04599.pdf

    Expected calibration error is calculated by dividing predictions into buckets based on the model's confidence (a probability value between 0 and 1).
    We then calculate the accuracy within each bucket and calculate the average gap between confidence and accuracy
    across buckets, weighted by the number of samples in each bucket.

    Each task must implement its own definition of "confidence" to be computed via the `update` method.

    Adds metric state variables:
    bucket_totals (float): The number of instances where the prediction masked the target per bucket.
    bucket_correct (float): The number of total instances that were predicted per bucket.

    Args:
        dist_sync_on_step (bool, optional): Synchronize metric state across processes at
            each forward() before returning the value at the step. Default: ``False``.
        n_buckets (int): Number of distinct buckets to split the confidence distribution into
    """

    def __init__(self, dist_sync_on_step: bool = False, n_buckets: int = 10):
        # state from multiple processes
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.n_buckets = n_buckets
        if n_buckets < 1:
            raise Exception('`n_buckets`')
        self.add_state(
            'bucket_totals',
            default=torch.zeros(n_buckets),
            dist_reduce_fx='sum',
        )
        self.add_state(
            'bucket_correct',
            default=torch.zeros(n_buckets),
            dist_reduce_fx='sum',
        )

    def update(self, batch: dict, outputs: torch.Tensor, labels: torch.Tensor):
        pass

    def compute(self):
        assert isinstance(self.bucket_correct, Tensor)
        assert isinstance(self.bucket_totals, Tensor)

        result = torch.tensor(0.0, device=self.bucket_correct.device)
        total_obs = torch.sum(self.bucket_totals)
        for i in range(self.n_buckets):
            if self.bucket_totals[i] == 0:
                continue

            acc_bucket_i = self.bucket_correct[i] / self.bucket_totals[i]
            upper_bound = (i + 1) / self.n_buckets
            lower_bound = i / self.n_buckets
            conf_bucket_i = torch.tensor((upper_bound + lower_bound) / 2,
                                         device=self.bucket_correct.device)
            result += (self.bucket_totals[i] /
                       total_obs) * torch.abs(acc_bucket_i - conf_bucket_i)
        return result


class InContextLearningMCExpectedCalibrationError(
    InContextLearningExpectedCalibrationError,
):
    r"""Computes Expected Calibration Error (ECE) for In-context learning (ICL)

    multiple choice (MC) tasks. (source: https://arxiv.org/abs/2012.00955).

    For MC tasks, the model confidence is defined as the softmax of average per-token probability assigned to the top question choice.

    See `InContextLearningExpectedCalibrationError` for more info.
    """

    # Make torchmetrics call update only once
    full_state_update = False

    def update(self, batch: dict, outputs: torch.Tensor, labels: torch.Tensor):

        outputs = torch.softmax(outputs, dim=2)
        probabilities = []
        for batch_idx, cont_idx in enumerate(batch['continuation_indices']):
            cont_tok_logits = outputs[batch_idx].index_select(
                dim=0,
                index=cont_idx - 1,
            )
            cont_tok_targ = labels[batch_idx].index_select(
                dim=0,
                index=cont_idx - 1,
            )
            probability = cont_tok_logits.index_select(
                dim=1,
                index=cont_tok_targ,
            ).diagonal().mean()
            probabilities.append(probability)

        for (start, end), gold_idx in zip(
            batch['choice_groupings'],
            batch['gold_indices'],
        ):
            subset = probabilities[start:end]
            idx_max = subset.index(max(subset))
            confidence = torch.tensor(subset).max() / torch.tensor(subset).sum()

            assert confidence >= 0.0 and confidence <= 1.0
            bucket_idx = int(confidence * self.n_buckets)
            if bucket_idx == self.n_buckets:
                bucket_idx -= 1

            if idx_max == gold_idx:
                self.bucket_correct[
                    bucket_idx
                ] += 1  # pyright: ignore [reportGeneralTypeIssues]

            self.bucket_totals[
                bucket_idx] += 1  # pyright: ignore [reportGeneralTypeIssues]


class InContextLearningLMExpectedCalibrationError(
    InContextLearningExpectedCalibrationError,
):
    r"""Computes Expected Calibration Error (ECE) for In-context learning (ICL)

    language modeling (LM) tasks. (cite: https://arxiv.org/pdf/1706.04599.pdf).

    For LM tasks, the model confidence is defined as the minimum probability assigned to all tokens in the continuation.

    See `InContextLearningExpectedCalibrationError` for more info.
    """

    # Make torchmetrics call update only once
    full_state_update = False

    def update(self, batch: dict, outputs: torch.Tensor, labels: torch.Tensor):

        outputs = torch.softmax(outputs, dim=2)
        for batch_idx, cont_idx in enumerate(batch['continuation_indices']):
            cont_tok_logits = outputs[batch_idx].index_select(
                dim=0,
                index=cont_idx - 1,
            )
            cont_tok_pred = cont_tok_logits.argmax(dim=-1)
            confidence = cont_tok_logits.max(dim=-1).values.min()
            cont_tok_targ = labels[batch_idx].index_select(
                dim=0,
                index=cont_idx - 1,
            )
            assert confidence >= 0.0 and confidence <= 1.0
            bucket_idx = int(confidence * self.n_buckets)
            if bucket_idx == self.n_buckets:
                bucket_idx -= 1

            if (cont_tok_pred == cont_tok_targ).all():
                self.bucket_correct[
                    bucket_idx
                ] += 1  # pyright: ignore [reportGeneralTypeIssues]

            self.bucket_totals[
                bucket_idx] += 1  # pyright: ignore [reportGeneralTypeIssues]



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry.registry import tp_strategies
from llmfoundry.tp.ffn_tp_strategy import ffn_tp_strategy

tp_strategies.register('ffn', func=ffn_tp_strategy)

__all__ = [
    'ffn_tp_strategy',
]
